"""Tests for common loader functions (used across multiple formats)."""

from datetime import date, datetime, time

import pytest

from dature.sources_loader.loaders.common import (
    bool_loader,
    bytearray_from_json_string,
    bytearray_from_string,
    date_from_string,
    date_passthrough,
    datetime_from_string,
    datetime_passthrough,
    int_from_string,
    none_from_empty_string,
    optional_from_empty_string,
    time_from_string,
)

# === Date/Time converters ===


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("2024-12-31", date(2024, 12, 31)),
    ],
)
def test_date_from_string(input_value, expected):
    assert date_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("2024-12-31T23:59:59", datetime(2024, 12, 31, 23, 59, 59)),
    ],
)
def test_datetime_from_string(input_value, expected):
    assert datetime_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("10:30:45", time(10, 30, 45)),
        ("10:30", time(10, 30)),
    ],
)
def test_time_from_string(input_value, expected):
    assert time_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        (date(2024, 12, 31), date(2024, 12, 31)),
    ],
)
def test_date_passthrough(input_value, expected):
    assert date_passthrough(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        (datetime(2024, 12, 31, 23, 59, 59), datetime(2024, 12, 31, 23, 59, 59)),
    ],
)
def test_datetime_passthrough(input_value, expected):
    assert datetime_passthrough(input_value) == expected


# === string converters ===


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("hello", bytearray(b"hello")),
    ],
)
def test_bytearray_from_string(input_value, expected):
    assert bytearray_from_string(input_value) == expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("", None),
    ],
)
def test_none_from_empty_string(input_value, expected):
    assert none_from_empty_string(input_value) is expected


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("", None),
        ("some text", "some text"),
    ],
)
def test_optional_from_empty_string(input_value, expected):
    assert optional_from_empty_string(input_value) == expected


# === Bool converter ===


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("true", True),
        ("TRUE", True),
        ("1", True),
        ("yes", True),
        ("on", True),
        ("false", False),
        ("FALSE", False),
        ("0", False),
        ("no", False),
        ("off", False),
        ("", False),
        (True, True),
        (False, False),
    ],
)
def test_bool_loader(input_value, expected):
    assert bool_loader(input_value) is expected


# === Int converter ===


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        (42, 42),
        ("42", 42),
        ("-1", -1),
        ("0", 0),
    ],
)
def test_int_from_string(input_value, expected):
    assert int_from_string(input_value) == expected


@pytest.mark.parametrize(
    "input_value",
    [True, False],
)
def test_int_from_string_rejects_bool(input_value):
    with pytest.raises(TypeError, match="Expected int, got bool"):
        int_from_string(input_value)


# === JSON string converters ===


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        ("hello", bytearray(b"hello")),
        ("", bytearray()),
    ],
)
def test_bytearray_from_json_string(input_value, expected):
    assert bytearray_from_json_string(input_value) == expected



from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Literal

import pytest

from dature import LoadMetadata, load
from dature.errors.exceptions import DatureConfigError, FieldLoadError
from dature.sources_loader.yaml_ import Yaml11Loader, Yaml12Loader
from dature.validators.number import Ge, Le
from dature.validators.sequence import MinItems, UniqueItems
from dature.validators.string import MaxLength, MinLength, RegexPattern


@dataclass
class Address:
    city: Annotated[str, MinLength(value=2)]
    zip_code: Annotated[str, RegexPattern(pattern=r"^\d{5}$")]


@dataclass
class ErrorConfig:
    port: int
    host: str
    status: Literal["active", "inactive"]
    name: Annotated[str, MinLength(value=3), MaxLength(value=50)]
    email: Annotated[str, RegexPattern(pattern=r"^[\w.-]+@[\w.-]+\.\w+$")]
    age: Annotated[int, Ge(value=0), Le(value=150)]
    tags: Annotated[list[str], MinItems(value=1), UniqueItems()]
    address: Address


@dataclass
class LoadErrorConfig:
    port: int
    host: str
    status: Literal["active", "inactive"]
    name: str
    email: str
    age: int
    tags: list[str]
    address: Address


@dataclass
class ValidationErrorConfig:
    name: Annotated[str, MinLength(value=3), MaxLength(value=50)]
    email: Annotated[str, RegexPattern(pattern=r"^[\w.-]+@[\w.-]+\.\w+$")]
    age: Annotated[int, Ge(value=0), Le(value=150)]
    tags: Annotated[list[str], MinItems(value=1), UniqueItems()]
    address: Address


FIXTURES_DIR = Path(__file__).parent.parent / "fixtures"

ALL_SOURCES = [
    ("errors.json", {}),
    ("errors.json5", {}),
    ("errors.yaml", {}),
    ("errors.yaml", {"loader": Yaml11Loader}),
    ("errors.yaml", {"loader": Yaml12Loader}),
    ("errors.toml", {}),
    ("errors.ini", {"prefix": "config"}),
    ("errors.env", {}),
]

EXPECTED_LOAD_ERRORS = [
    (["port"], "invalid literal for int() with base 10: 'abc'"),
    (["host"], "Missing required field"),
    (["status"], "Invalid variant: 'unknown'"),
    (["address", "city"], "Value must have at least 2 characters"),
    (["address", "zip_code"], r"Value must match pattern '^\d{5}$'"),
]

EXPECTED_VALIDATION_ERRORS = [
    (["name"], "Value must have at least 3 characters"),
    (["email"], r"Value must match pattern '^[\w.-]+@[\w.-]+\.\w+$'"),
    (["age"], "Value must be less than or equal to 150"),
    (["tags"], "Value must have at least 1 items"),
    (["address", "city"], "Value must have at least 2 characters"),
    (["address", "zip_code"], r"Value must match pattern '^\d{5}$'"),
]


def _assert_field_errors(
    exceptions: tuple[Exception, ...],
    expected: list[tuple[list[str], str]],
) -> None:
    assert len(exceptions) == len(expected)
    for exc, (path, message) in zip(exceptions, expected, strict=True):
        assert isinstance(exc, FieldLoadError)
        assert exc.field_path == path
        assert exc.message == message


@pytest.mark.parametrize(("fixture_file", "metadata_kwargs"), ALL_SOURCES)
def test_load_error_types(
    fixture_file: str,
    metadata_kwargs: dict[str, str],
) -> None:
    metadata = LoadMetadata(file_=str(FIXTURES_DIR / fixture_file), **metadata_kwargs)

    with pytest.raises(DatureConfigError) as exc_info:
        load(metadata, LoadErrorConfig)

    err = exc_info.value
    assert str(err) == f"LoadErrorConfig loading errors ({len(EXPECTED_LOAD_ERRORS)})"
    _assert_field_errors(err.exceptions, EXPECTED_LOAD_ERRORS)


@pytest.mark.parametrize(("fixture_file", "metadata_kwargs"), ALL_SOURCES)
def test_validation_error_types(
    fixture_file: str,
    metadata_kwargs: dict[str, str],
) -> None:
    metadata = LoadMetadata(file_=str(FIXTURES_DIR / fixture_file), **metadata_kwargs)

    with pytest.raises(DatureConfigError) as exc_info:
        load(metadata, ValidationErrorConfig)

    err = exc_info.value
    assert str(err) == f"ValidationErrorConfig loading errors ({len(EXPECTED_VALIDATION_ERRORS)})"
    _assert_field_errors(err.exceptions, EXPECTED_VALIDATION_ERRORS)



"""Load from file-like objects (BytesIO, StringIO)."""

from dataclasses import dataclass
from io import BytesIO, StringIO

from dature import LoadMetadata, load
from dature.sources_loader.json_ import JsonLoader


@dataclass
class Config:
    host: str
    port: int
    debug: bool = False


# From StringIO
text_stream = StringIO('{"host": "localhost", "port": 8080, "debug": true}')
config = load(LoadMetadata(file_=text_stream, loader=JsonLoader), Config)

assert config.host == "localhost"
assert config.port == 8080

# From BytesIO
binary_stream = BytesIO(b'{"host": "0.0.0.0", "port": 3000}')
config = load(LoadMetadata(file_=binary_stream, loader=JsonLoader), Config)

assert config.host == "0.0.0.0"
assert config.port == 3000



from dature.errors.exceptions import LineRange
from dature.path_finders.json_ import JsonPathFinder


class TestJsonFindLineRange:
    def test_key_not_confused_by_newline_in_value(self):
        content = '{\n  "str1": "line1\\nkey=1",\n  "key": "real"\n}'
        finder = JsonPathFinder(content)

        assert finder.find_line_range(["key"]) == LineRange(start=3, end=3)

    def test_key_not_confused_by_escaped_quotes(self):
        content = '{\n  "str1": "he said \\"key\\": value",\n  "key": 42\n}'
        finder = JsonPathFinder(content)

        assert finder.find_line_range(["key"]) == LineRange(start=3, end=3)

    def test_scalar_value(self):
        content = '{\n  "timeout": 30\n}'
        finder = JsonPathFinder(content)

        assert finder.find_line_range(["timeout"]) == LineRange(start=2, end=2)

    def test_multiline_dict(self):
        content = '{\n  "db": {\n    "host": "localhost",\n    "port": 5432\n  }\n}'
        finder = JsonPathFinder(content)

        assert finder.find_line_range(["db"]) == LineRange(start=2, end=5)

    def test_multiline_array(self):
        content = '{\n  "tags": [\n    "a",\n    "b"\n  ]\n}'
        finder = JsonPathFinder(content)

        assert finder.find_line_range(["tags"]) == LineRange(start=2, end=5)

    def test_not_found(self):
        content = '{\n  "name": "test"\n}'
        finder = JsonPathFinder(content)

        assert finder.find_line_range(["missing"]) is None

    def test_inline_dict(self):
        content = '{\n  "db": {"host": "localhost"}\n}'
        finder = JsonPathFinder(content)

        assert finder.find_line_range(["db"]) == LineRange(start=2, end=2)

    def test_inline_array(self):
        content = '{\n  "tags": ["a", "b"]\n}'
        finder = JsonPathFinder(content)

        assert finder.find_line_range(["tags"]) == LineRange(start=2, end=2)



"""Tests for json5_ loader functions."""

import pytest
from json5 import JsonIdentifier

from dature.sources_loader.loaders.json5_ import str_from_json_identifier


@pytest.mark.parametrize(
    ("input_value", "expected"),
    [
        (JsonIdentifier(""), ""),
        (JsonIdentifier("some_key"), "some_key"),
    ],
)
def test_str_from_json_identifier(input_value: JsonIdentifier, expected: str) -> None:
    assert str_from_json_identifier(input_value) == expected



"""Tuple shorthand as a decorator — implicit LAST_WINS merge."""

import os
from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, load

SHARED_DIR = Path(__file__).parents[2] / "shared"

os.environ["APP_HOST"] = "env_localhost"


@load(
    (
        LoadMetadata(file_=SHARED_DIR / "common_defaults.yaml"),
        LoadMetadata(prefix="APP_"),
    ),
)
@dataclass
class Config:
    host: str
    port: int
    debug: bool = False


config = Config()
assert config.host == "env_localhost"
assert config.port == 3000
assert config.debug is False



from dataclasses import dataclass

from dature.errors.exceptions import MergeConflictError, MergeConflictFieldError, SourceLocation
from dature.errors.location import resolve_source_location
from dature.loading.source_loading import SourceContext
from dature.metadata import FieldMergeStrategy, MergeStrategy
from dature.types import JSONValue

_MIN_CONFLICT_SOURCES = 2


def _deduplicate_list(items: list[JSONValue]) -> list[JSONValue]:
    seen: list[JSONValue] = []
    for item in items:
        found = False
        for s in seen:
            if s == item:
                found = True
                break
        if not found:
            seen.append(item)
    return seen


@dataclass(frozen=True, slots=True)
class ListPair:
    base: list[JSONValue]
    override: list[JSONValue]


def _ensure_both_lists(
    base: JSONValue,
    override: JSONValue,
    strategy_name: str,
) -> ListPair:
    if not isinstance(base, list) or not isinstance(override, list):
        base_type = type(base).__name__
        override_type = type(override).__name__
        msg = f"{strategy_name} strategy requires both values to be lists, got {base_type} and {override_type}"
        raise TypeError(msg)
    return ListPair(base=base, override=override)


def _apply_list_merge(
    base: JSONValue,
    override: JSONValue,
    strategy: FieldMergeStrategy,
) -> list[JSONValue]:
    if strategy == FieldMergeStrategy.APPEND:
        pair = _ensure_both_lists(base, override, "APPEND")
        return list(pair.base) + list(pair.override)

    if strategy == FieldMergeStrategy.APPEND_UNIQUE:
        pair = _ensure_both_lists(base, override, "APPEND_UNIQUE")
        return _deduplicate_list(list(pair.base) + list(pair.override))

    if strategy == FieldMergeStrategy.PREPEND:
        pair = _ensure_both_lists(base, override, "PREPEND")
        return list(pair.override) + list(pair.base)

    # PREPEND_UNIQUE
    pair = _ensure_both_lists(base, override, "PREPEND_UNIQUE")
    return _deduplicate_list(list(pair.override) + list(pair.base))


def apply_field_merge(
    base: JSONValue,
    override: JSONValue,
    strategy: FieldMergeStrategy,
) -> JSONValue:
    if strategy == FieldMergeStrategy.FIRST_WINS:
        return base

    if strategy == FieldMergeStrategy.LAST_WINS:
        return override

    return _apply_list_merge(base, override, strategy)


def deep_merge_last_wins(
    base: JSONValue,
    override: JSONValue,
    *,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
    _path: str = "",
) -> JSONValue:
    if field_merge_map is not None and _path in field_merge_map:
        return apply_field_merge(base, override, field_merge_map[_path])

    if isinstance(base, dict) and isinstance(override, dict):
        result = dict(base)
        for key, value in override.items():
            child_path = f"{_path}.{key}" if _path else key
            if key in result:
                result[key] = deep_merge_last_wins(
                    result[key],
                    value,
                    field_merge_map=field_merge_map,
                    _path=child_path,
                )
            else:
                result[key] = value
        return result
    return override


def deep_merge_first_wins(
    base: JSONValue,
    override: JSONValue,
    *,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
    _path: str = "",
) -> JSONValue:
    if field_merge_map is not None and _path in field_merge_map:
        return apply_field_merge(base, override, field_merge_map[_path])

    if isinstance(base, dict) and isinstance(override, dict):
        result = dict(base)
        for key, value in override.items():
            child_path = f"{_path}.{key}" if _path else key
            if key in result:
                result[key] = deep_merge_first_wins(
                    result[key],
                    value,
                    field_merge_map=field_merge_map,
                    _path=child_path,
                )
            else:
                result[key] = value
        return result
    return base


def _collect_conflicts(
    dicts: list[JSONValue],
    source_contexts: list[SourceContext],
    path: list[str],
    conflicts: list[tuple[list[str], list[tuple[int, JSONValue]]]],
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
    callable_merge_paths: frozenset[str] | None = None,
) -> None:
    key_sources: dict[str, list[tuple[int, JSONValue]]] = {}

    for i, d in enumerate(dicts):
        if not isinstance(d, dict):
            continue
        for key, value in d.items():
            if key not in key_sources:
                key_sources[key] = []
            key_sources[key].append((i, value))

    for key, sources in key_sources.items():
        if len(sources) < _MIN_CONFLICT_SOURCES:
            continue

        field_path = ".".join([*path, key])
        has_enum_strategy = field_merge_map is not None and field_path in field_merge_map
        has_callable_strategy = callable_merge_paths is not None and field_path in callable_merge_paths
        if has_enum_strategy or has_callable_strategy:
            continue

        values = [v for _, v in sources]

        nested_dicts = [v for v in values if isinstance(v, dict)]
        if len(nested_dicts) == len(sources):
            _collect_conflicts(
                values,
                [source_contexts[i] for i, _ in sources],
                [*path, key],
                conflicts,
                field_merge_map=field_merge_map,
                callable_merge_paths=callable_merge_paths,
            )
            continue

        if all(v == values[0] for v in values[1:]):
            continue

        conflicts.append(([*path, key], sources))


def raise_on_conflict(
    dicts: list[JSONValue],
    source_ctxs: list[SourceContext],
    dataclass_name: str,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
    callable_merge_paths: frozenset[str] | None = None,
) -> None:
    conflicts: list[tuple[list[str], list[tuple[int, JSONValue]]]] = []
    _collect_conflicts(
        dicts,
        source_ctxs,
        [],
        conflicts,
        field_merge_map=field_merge_map,
        callable_merge_paths=callable_merge_paths,
    )

    if not conflicts:
        return

    conflict_errors: list[MergeConflictFieldError] = []
    for field_path, sources in conflicts:
        locations: list[SourceLocation] = []
        for source_idx, _ in sources:
            source_ctx = source_ctxs[source_idx]
            locs = resolve_source_location(field_path, source_ctx.error_ctx, source_ctx.file_content)
            locations.extend(locs)
        conflict_errors.append(
            MergeConflictFieldError(
                field_path=field_path,
                message="Conflicting values in multiple sources",
                locations=locations,
            ),
        )

    raise MergeConflictError(dataclass_name, conflict_errors)


def deep_merge(
    base: JSONValue,
    override: JSONValue,
    *,
    strategy: MergeStrategy,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
) -> JSONValue:
    if strategy == MergeStrategy.LAST_WINS:
        return deep_merge_last_wins(base, override, field_merge_map=field_merge_map)
    if strategy in (MergeStrategy.FIRST_WINS, MergeStrategy.FIRST_FOUND):
        return deep_merge_first_wins(base, override, field_merge_map=field_merge_map)
    msg = "Use merge_sources for RAISE_ON_CONFLICT strategy"
    raise ValueError(msg)



import logging
from collections.abc import Callable
from dataclasses import dataclass as stdlib_dataclass
from dataclasses import fields, is_dataclass
from typing import Any

from dature.config import config
from dature.errors.exceptions import DatureConfigError
from dature.errors.formatter import enrich_skipped_errors, handle_load_errors
from dature.load_report import (
    FieldOrigin,
    LoadReport,
    SourceEntry,
    attach_load_report,
    compute_field_origins,
    get_load_report,
)
from dature.loading.context import (
    build_error_ctx,
    coerce_flag_fields,
    ensure_retort,
    make_validating_post_init,
    merge_fields,
)
from dature.loading.resolver import resolve_loader
from dature.loading.source_loading import load_sources, resolve_expand_env_vars
from dature.masking.detection import build_secret_paths
from dature.masking.masking import (
    mask_field_origins,
    mask_json_value,
    mask_source_entries,
    mask_value,
)
from dature.merging.deep_merge import deep_merge, deep_merge_last_wins, raise_on_conflict
from dature.merging.field_group import FieldGroupContext, validate_field_groups
from dature.merging.predicate import ResolvedFieldGroup, build_field_group_paths, build_field_merge_map
from dature.metadata import FieldMergeStrategy, LoadMetadata, MergeMetadata, MergeStrategy, TypeLoader
from dature.protocols import DataclassInstance, LoaderProtocol
from dature.types import FieldMergeCallable, JSONValue

logger = logging.getLogger("dature")


def _resolve_merge_mask_secrets(merge_meta: MergeMetadata) -> bool:
    if merge_meta.mask_secrets is not None:
        return merge_meta.mask_secrets
    return config.masking.mask_secrets


def _collect_extra_secret_patterns(merge_meta: MergeMetadata) -> tuple[str, ...]:
    merge_names = merge_meta.secret_field_names or ()
    source_names: list[str] = []
    for source_meta in merge_meta.sources:
        if source_meta.secret_field_names is not None:
            source_names.extend(source_meta.secret_field_names)
    return merge_names + tuple(source_names)


def _log_merge_step(  # noqa: PLR0913
    *,
    dataclass_name: str,
    step_idx: int,
    strategy: MergeStrategy,
    before: JSONValue,
    source_data: JSONValue,
    after: JSONValue,
    secret_paths: frozenset[str] = frozenset(),
) -> None:
    if isinstance(before, dict) and isinstance(source_data, dict) and isinstance(after, dict):
        added_keys = set(source_data.keys()) - set(before.keys())
        overwritten_keys = set(source_data.keys()) & set(before.keys())
        logger.debug(
            "[%s] Merge step %d (strategy=%s): added=%s, overwritten=%s",
            dataclass_name,
            step_idx,
            strategy.value,
            sorted(added_keys),
            sorted(overwritten_keys),
        )
    if secret_paths:
        masked_after = mask_json_value(after, secret_paths=secret_paths)
    else:
        masked_after = after
    logger.debug(
        "[%s] State after step %d: %s",
        dataclass_name,
        step_idx,
        masked_after,
    )


def _log_field_origins(
    *,
    dataclass_name: str,
    field_origins: tuple[FieldOrigin, ...],
    secret_paths: frozenset[str] = frozenset(),
) -> None:
    for origin in field_origins:
        if origin.key in secret_paths:
            masked = mask_value(str(origin.value))
            logger.debug(
                "[%s] Field '%s' = %r  <-- source %d (%s)",
                dataclass_name,
                origin.key,
                masked,
                origin.source_index,
                origin.source_file or "<env>",
            )
        else:
            logger.debug(
                "[%s] Field '%s' = %r  <-- source %d (%s)",
                dataclass_name,
                origin.key,
                origin.value,
                origin.source_index,
                origin.source_file or "<env>",
            )


def _build_merge_report(
    *,
    dataclass_name: str,
    strategy: MergeStrategy,
    source_entries: tuple[SourceEntry, ...],
    field_origins: tuple[FieldOrigin, ...],
    merged_data: JSONValue,
    secret_paths: frozenset[str] = frozenset(),
) -> LoadReport:
    if secret_paths:
        source_entries = mask_source_entries(source_entries, secret_paths=secret_paths)
        field_origins = mask_field_origins(field_origins, secret_paths=secret_paths)
        merged_data = mask_json_value(merged_data, secret_paths=secret_paths)

    return LoadReport(
        dataclass_name=dataclass_name,
        strategy=strategy,
        sources=source_entries,
        field_origins=field_origins,
        merged_data=merged_data,
    )


def _collect_leaf_paths(data: JSONValue, prefix: str = "") -> list[str]:
    if not isinstance(data, dict):
        return [prefix] if prefix else []
    paths: list[str] = []
    for key, value in data.items():
        child_path = f"{prefix}.{key}" if prefix else key
        if isinstance(value, dict):
            paths.extend(_collect_leaf_paths(value, child_path))
        else:
            paths.append(child_path)
    return paths


def _validate_all_field_groups(
    *,
    raw_dicts: list[JSONValue],
    field_group_paths: tuple[ResolvedFieldGroup, ...],
    dataclass_name: str,
    source_reprs: tuple[str, ...],
) -> None:
    merged: JSONValue = {}
    field_origins: dict[str, int] = {}
    ctx = FieldGroupContext(
        source_reprs=source_reprs,
        field_origins=field_origins,
        dataclass_name=dataclass_name,
    )
    for step_idx, raw in enumerate(raw_dicts):
        validate_field_groups(
            base=merged,
            source=raw,
            field_group_paths=field_group_paths,
            source_index=step_idx,
            ctx=ctx,
        )
        for leaf_path in _collect_leaf_paths(raw):
            field_origins[leaf_path] = step_idx
        merged = deep_merge_last_wins(merged, raw, field_merge_map=None)


def _collect_field_values(
    raw_dicts: list[JSONValue],
    field_path: str,
) -> list[JSONValue]:
    parts = field_path.split(".")
    values: list[JSONValue] = []
    for raw in raw_dicts:
        current: JSONValue = raw
        found = True
        for part in parts:
            if not isinstance(current, dict) or part not in current:
                found = False
                break
            current = current[part]
        if found:
            values.append(current)
    return values


def _set_nested_value(
    data: JSONValue,
    field_path: str,
    value: JSONValue,
) -> JSONValue:
    if not isinstance(data, dict):
        return data
    parts = field_path.split(".")
    if len(parts) == 1:
        result = dict(data)
        result[parts[0]] = value
        return result
    key = parts[0]
    rest = ".".join(parts[1:])
    result = dict(data)
    if key in result:
        result[key] = _set_nested_value(result[key], rest, value)
    return result


def _merge_raw_dicts(
    *,
    raw_dicts: list[JSONValue],
    strategy: MergeStrategy,
    dataclass_name: str,
    field_merge_map: dict[str, FieldMergeStrategy] | None = None,
    callable_merge_map: dict[str, FieldMergeCallable] | None = None,
    secret_paths: frozenset[str] = frozenset(),
) -> JSONValue:
    merged: JSONValue = {}
    for step_idx, raw in enumerate(raw_dicts):
        before = merged

        if strategy == MergeStrategy.RAISE_ON_CONFLICT:
            merged = deep_merge_last_wins(merged, raw, field_merge_map=field_merge_map)
        else:
            merged = deep_merge(merged, raw, strategy=strategy, field_merge_map=field_merge_map)

        _log_merge_step(
            dataclass_name=dataclass_name,
            step_idx=step_idx,
            strategy=strategy,
            before=before,
            source_data=raw,
            after=merged,
            secret_paths=secret_paths,
        )

    if callable_merge_map:
        for field_path, merge_fn in callable_merge_map.items():
            values = _collect_field_values(raw_dicts, field_path)
            if not values:
                continue
            aggregated = merge_fn(values)
            merged = _set_nested_value(merged, field_path, aggregated)

    return merged


@stdlib_dataclass(frozen=True, slots=True)
class _MergedData[T: DataclassInstance]:
    result: T
    merged_raw: JSONValue
    last_loader: LoaderProtocol
    last_source_meta: LoadMetadata


def _load_and_merge[T: DataclassInstance](  # noqa: C901
    *,
    merge_meta: MergeMetadata,
    dataclass_: type[T],
    loaders: tuple[LoaderProtocol, ...] | None = None,
    debug: bool = False,
    type_loaders: tuple[TypeLoader, ...] = (),
) -> _MergedData[T]:
    secret_paths: frozenset[str] = frozenset()
    if _resolve_merge_mask_secrets(merge_meta):
        extra_patterns = _collect_extra_secret_patterns(merge_meta)
        secret_paths = build_secret_paths(dataclass_, extra_patterns=extra_patterns)

    loaded = load_sources(
        merge_meta=merge_meta,
        dataclass_name=dataclass_.__name__,
        dataclass_=dataclass_,
        loaders=loaders,
        secret_paths=secret_paths,
        mask_secrets=_resolve_merge_mask_secrets(merge_meta),
        type_loaders=type_loaders,
    )

    merge_maps = build_field_merge_map(merge_meta.field_merges, dataclass_)

    field_group_paths: tuple[ResolvedFieldGroup, ...] = ()
    if merge_meta.field_groups:
        field_group_paths = build_field_group_paths(merge_meta.field_groups, dataclass_)

    if field_group_paths:
        source_reprs = tuple(repr(merge_meta.sources[entry.index]) for entry in loaded.source_entries)
        _validate_all_field_groups(
            raw_dicts=loaded.raw_dicts,
            field_group_paths=field_group_paths,
            dataclass_name=dataclass_.__name__,
            source_reprs=source_reprs,
        )

    if merge_meta.strategy == MergeStrategy.RAISE_ON_CONFLICT:
        raise_on_conflict(
            loaded.raw_dicts,
            loaded.source_ctxs,
            dataclass_.__name__,
            field_merge_map=merge_maps.enum_map or None,
            callable_merge_paths=merge_maps.callable_paths or None,
        )

    merged = _merge_raw_dicts(
        raw_dicts=loaded.raw_dicts,
        strategy=merge_meta.strategy,
        dataclass_name=dataclass_.__name__,
        field_merge_map=merge_maps.enum_map or None,
        callable_merge_map=merge_maps.callable_map or None,
        secret_paths=secret_paths,
    )

    if secret_paths:
        masked_merged = mask_json_value(merged, secret_paths=secret_paths)
    else:
        masked_merged = merged
    logger.debug(
        "[%s] Merged result (strategy=%s, %d sources): %s",
        dataclass_.__name__,
        merge_meta.strategy.value,
        len(loaded.raw_dicts),
        masked_merged,
    )

    frozen_entries = tuple(loaded.source_entries)
    field_origins = compute_field_origins(
        raw_dicts=loaded.raw_dicts,
        source_entries=frozen_entries,
        strategy=merge_meta.strategy,
    )

    _log_field_origins(
        dataclass_name=dataclass_.__name__,
        field_origins=field_origins,
        secret_paths=secret_paths,
    )

    report: LoadReport | None = None
    if debug:
        report = _build_merge_report(
            dataclass_name=dataclass_.__name__,
            strategy=merge_meta.strategy,
            source_entries=frozen_entries,
            field_origins=field_origins,
            merged_data=merged,
            secret_paths=secret_paths,
        )

    last_error_ctx = loaded.source_ctxs[-1].error_ctx
    merged = coerce_flag_fields(merged, dataclass_)
    try:
        result = handle_load_errors(
            func=lambda: loaded.last_loader.transform_to_dataclass(merged, dataclass_),
            ctx=last_error_ctx,
        )
    except DatureConfigError as exc:
        if report is not None:
            attach_load_report(dataclass_, report)
        if loaded.skipped_fields:
            raise enrich_skipped_errors(exc, loaded.skipped_fields) from exc
        raise

    if report is not None:
        attach_load_report(result, report)

    last_source_idx = loaded.source_entries[-1].index
    return _MergedData(
        result=result,
        merged_raw=merged,
        last_loader=loaded.last_loader,
        last_source_meta=merge_meta.sources[last_source_idx],
    )


def merge_load_as_function[T: DataclassInstance](
    merge_meta: MergeMetadata,
    dataclass_: type[T],
    *,
    debug: bool,
    type_loaders: tuple[TypeLoader, ...] = (),
) -> T:
    data = _load_and_merge(
        merge_meta=merge_meta,
        dataclass_=dataclass_,
        debug=debug,
        type_loaders=type_loaders,
    )

    validating_retort = data.last_loader.create_validating_retort(dataclass_)
    validation_loader = validating_retort.get_loader(dataclass_)

    last_meta = data.last_source_meta
    mask_secrets = _resolve_merge_mask_secrets(merge_meta)
    secret_paths: frozenset[str] = frozenset()
    if mask_secrets:
        extra_patterns = _collect_extra_secret_patterns(merge_meta)
        secret_paths = build_secret_paths(dataclass_, extra_patterns=extra_patterns)
    last_error_ctx = build_error_ctx(
        last_meta,
        dataclass_.__name__,
        secret_paths=secret_paths,
        mask_secrets=mask_secrets,
    )
    try:
        handle_load_errors(
            func=lambda: validation_loader(data.merged_raw),
            ctx=last_error_ctx,
        )
    except DatureConfigError:
        if debug:
            report = get_load_report(data.result)
            if report is not None:
                attach_load_report(dataclass_, report)
        raise

    return data.result


class _MergePatchContext:
    def __init__(
        self,
        *,
        merge_meta: MergeMetadata,
        cls: type[DataclassInstance],
        cache: bool,
        debug: bool,
        type_loaders: tuple[TypeLoader, ...] = (),
    ) -> None:
        self.loaders = self._prepare_loaders(merge_meta=merge_meta, cls=cls, type_loaders=type_loaders)

        self.merge_meta = merge_meta
        self.cls = cls
        self.cache = cache
        self.debug = debug
        self.type_loaders = type_loaders
        self.cached_data: DataclassInstance | None = None
        self.field_list = fields(cls)
        self.original_init = cls.__init__
        self.original_post_init = getattr(cls, "__post_init__", None)
        self.loading = False
        self.validating = False

        last_loader = self.loaders[-1]
        validating_retort = last_loader.create_validating_retort(cls)
        self.validation_loader: Callable[[JSONValue], DataclassInstance] = validating_retort.get_loader(cls)

        mask_secrets = _resolve_merge_mask_secrets(merge_meta)
        self.secret_paths: frozenset[str] = frozenset()
        if mask_secrets:
            extra_patterns = _collect_extra_secret_patterns(merge_meta)
            self.secret_paths = build_secret_paths(cls, extra_patterns=extra_patterns)

        last_meta = merge_meta.sources[-1]
        self.error_ctx = build_error_ctx(
            last_meta,
            cls.__name__,
            secret_paths=self.secret_paths,
            mask_secrets=mask_secrets,
        )

    @staticmethod
    def _prepare_loaders(
        *,
        merge_meta: MergeMetadata,
        cls: type[DataclassInstance],
        type_loaders: tuple[TypeLoader, ...] = (),
    ) -> tuple[LoaderProtocol, ...]:
        loaders: list[LoaderProtocol] = []
        for source_meta in merge_meta.sources:
            resolved_expand = resolve_expand_env_vars(source_meta, merge_meta)
            source_type_loaders = (source_meta.type_loaders or ()) + type_loaders
            resolved_strategy = (
                source_meta.nested_resolve_strategy
                or merge_meta.nested_resolve_strategy
                or config.loading.nested_resolve_strategy
            )
            resolve_kwargs: dict[str, Any] = {
                "expand_env_vars": resolved_expand,
                "type_loaders": source_type_loaders,
                "nested_resolve_strategy": resolved_strategy,
            }
            resolved_resolve = source_meta.nested_resolve or merge_meta.nested_resolve
            if resolved_resolve is not None:
                resolve_kwargs["nested_resolve"] = resolved_resolve
            loader_instance = resolve_loader(
                source_meta,
                **resolve_kwargs,
            )
            ensure_retort(loader_instance, cls)
            loaders.append(loader_instance)
        return tuple(loaders)


def _make_merge_new_init(ctx: _MergePatchContext) -> Callable[..., None]:
    def new_init(self: DataclassInstance, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        if ctx.loading:
            ctx.original_init(self, *args, **kwargs)
            return

        if ctx.cache and ctx.cached_data is not None:
            loaded_data = ctx.cached_data
        else:
            ctx.loading = True
            try:
                merged_data = _load_and_merge(
                    merge_meta=ctx.merge_meta,
                    dataclass_=ctx.cls,
                    loaders=ctx.loaders,
                    debug=ctx.debug,
                    type_loaders=ctx.type_loaders,
                )
            finally:
                ctx.loading = False
            loaded_data = merged_data.result
            ctx.error_ctx = build_error_ctx(
                merged_data.last_source_meta,
                ctx.cls.__name__,
                secret_paths=ctx.secret_paths,
                mask_secrets=ctx.error_ctx.mask_secrets,
            )
            if ctx.cache:
                ctx.cached_data = loaded_data

        complete_kwargs = merge_fields(loaded_data, ctx.field_list, args, kwargs)
        ctx.original_init(self, *args, **complete_kwargs)

        if ctx.debug:
            report = get_load_report(loaded_data)
            if report is not None:
                attach_load_report(self, report)

        if ctx.original_post_init is None:
            self.__post_init__()  # type: ignore[attr-defined]

    return new_init


def merge_make_decorator(
    merge_meta: MergeMetadata,
    *,
    cache: bool,
    debug: bool,
    type_loaders: tuple[TypeLoader, ...] = (),
) -> Callable[[type[DataclassInstance]], type[DataclassInstance]]:
    def decorator(cls: type[DataclassInstance]) -> type[DataclassInstance]:
        if not is_dataclass(cls):
            msg = f"{cls.__name__} must be a dataclass"
            raise TypeError(msg)

        ctx = _MergePatchContext(
            merge_meta=merge_meta,
            cls=cls,
            cache=cache,
            debug=debug,
            type_loaders=type_loaders,
        )
        cls.__init__ = _make_merge_new_init(ctx)  # type: ignore[method-assign]
        cls.__post_init__ = make_validating_post_init(ctx)  # type: ignore[attr-defined]
        return cls

    return decorator



import base64
import re
from datetime import timedelta
from urllib.parse import urlparse

from dature.fields.byte_size import ByteSize
from dature.fields.payment_card import PaymentCardNumber
from dature.fields.secret_str import SecretStr
from dature.types import URL, Base64UrlBytes, Base64UrlStr

_TIMEDELTA_RE = re.compile(
    r"^(?:(?P<weeks>-?\d+)\s+weeks?(?:,\s*|\s+|$))?"
    r"(?:(?P<days>-?\d+)\s+days?(?:,\s*|\s+|$))?"
    r"(?:(?P<sign>-)?(?P<hours>\d+):(?P<minutes>\d{2})(?::(?P<seconds>\d{2})(?:\.(?P<microseconds>\d+))?)?)?$",
)


def bytes_from_string(value: str) -> bytes:
    return value.encode("utf-8")


def complex_from_string(value: str) -> complex:
    return complex(value.replace(" ", ""))


def timedelta_from_string(value: str) -> timedelta:
    """Parse str(timedelta(...)) format: '1 day, 2:30:00', '2 days 1:02:03', '0:45:00', '2:30'."""
    match = _TIMEDELTA_RE.match(value)
    if match is None or value == "":
        msg = f"Invalid timedelta format: {value!r}"
        raise ValueError(msg)

    groups = match.groupdict()
    microseconds = 0
    if groups["microseconds"] is not None:
        microseconds = int(groups["microseconds"].ljust(6, "0"))

    sign = -1 if groups["sign"] is not None else 1

    return timedelta(
        weeks=int(groups["weeks"] or 0),
        days=int(groups["days"] or 0),
        hours=sign * int(groups["hours"] or 0),
        minutes=sign * int(groups["minutes"] or 0),
        seconds=sign * int(groups["seconds"] or 0),
        microseconds=sign * microseconds,
    )


def url_from_string(value: str) -> URL:
    return urlparse(value)


def base64url_bytes_from_string(value: str) -> Base64UrlBytes:
    return base64.urlsafe_b64decode(value)


def base64url_str_from_string(value: str) -> Base64UrlStr:
    return base64.urlsafe_b64decode(value).decode("utf-8")


def secret_str_from_string(value: str) -> SecretStr:
    return SecretStr(value)


def payment_card_number_from_string(value: str) -> PaymentCardNumber:
    return PaymentCardNumber(value)


def byte_size_from_string(value: str | int) -> ByteSize:
    return ByteSize(value)



"""Custom validator — error example."""

from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

from dature import LoadMetadata, load
from dature.errors.exceptions import DatureConfigError
from dature.validators.number import Ge

SOURCES_DIR = Path(__file__).parent / "sources"


@dataclass(frozen=True, slots=True, kw_only=True)
class Divisible:
    value: int
    error_message: str = "Value must be divisible by {value}"

    def get_validator_func(self) -> Callable[[int], bool]:
        def validate(val: int) -> bool:
            return val % self.value == 0

        return validate

    def get_error_message(self) -> str:
        return self.error_message.format(value=self.value)


@dataclass
class ServiceConfig:
    port: int
    name: str
    tags: list[str]
    workers: Annotated[int, Ge(value=1), Divisible(value=2)]


try:
    load(
        LoadMetadata(file_=SOURCES_DIR / "validation_custom_invalid.json5"),
        ServiceConfig,
    )
except DatureConfigError as exc:
    source = str(SOURCES_DIR / "validation_custom_invalid.json5")
    assert str(exc) == "ServiceConfig loading errors (1)"
    assert len(exc.exceptions) == 1
    assert str(exc.exceptions[0]) == (
        f"  [workers]  Value must be divisible by 2\n   └── FILE '{source}', line 5\n       workers: 3,"
    )



"""Annotated validators — error example."""

from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

from dature import LoadMetadata, load
from dature.errors.exceptions import DatureConfigError
from dature.validators.number import Ge, Le
from dature.validators.sequence import MinItems, UniqueItems
from dature.validators.string import MaxLength, MinLength

SOURCES_DIR = Path(__file__).parent / "sources"


@dataclass
class ServiceConfig:
    port: Annotated[int, Ge(value=1), Le(value=65535)]
    name: Annotated[str, MinLength(value=3), MaxLength(value=50)]
    tags: Annotated[list[str], MinItems(value=1), UniqueItems()]
    workers: Annotated[int, Ge(value=1)]


try:
    load(
        LoadMetadata(file_=SOURCES_DIR / "validation_annotated_invalid.json5"),
        ServiceConfig,
    )
except DatureConfigError as exc:
    source = str(SOURCES_DIR / "validation_annotated_invalid.json5")
    assert str(exc) == "ServiceConfig loading errors (4)"
    assert len(exc.exceptions) == 4
    assert str(exc.exceptions[0]) == (
        f"  [port]  Value must be greater than or equal to 1\n   └── FILE '{source}', line 3\n       port: 0,"
    )
    assert str(exc.exceptions[1]) == (
        f"  [name]  Value must have at least 3 characters\n   └── FILE '{source}', line 4\n       name: \"ab\","
    )
    assert str(exc.exceptions[2]) == (
        f'  [tags]  Value must contain unique items\n   └── FILE \'{source}\', line 5\n       tags: ["web", "web"],'
    )
    assert str(exc.exceptions[3]) == (
        f"  [workers]  Value must be greater than or equal to 1\n   └── FILE '{source}', line 6\n       workers: 0,"
    )



"""The problem: both JSON and flat keys exist for the same nested field."""

import os
from dataclasses import dataclass

from dature import LoadMetadata, load
from dature.sources_loader.env_ import EnvLoader

os.environ["APP__DATABASE"] = '{"host": "json-host", "port": "5432"}'
os.environ["APP__DATABASE__HOST"] = "flat-host"
os.environ["APP__DATABASE__PORT"] = "3306"


@dataclass
class Database:
    host: str
    port: int


@dataclass
class Config:
    database: Database


# Without nested_resolve_strategy, flat keys win by default
config = load(LoadMetadata(loader=EnvLoader, prefix="APP__"), Config)

assert config.database.host == "flat-host"
assert config.database.port == 3306



from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import pytest

from dature import LoadMetadata, load
from dature.errors.exceptions import DatureConfigError


@dataclass(frozen=True, slots=True, kw_only=True)
class Divisible:
    value: int
    error_message: str = "Value must be divisible by {value}"

    def get_validator_func(self) -> Callable[[int], bool]:
        def validate(val: int) -> bool:
            return val % self.value == 0

        return validate

    def get_error_message(self) -> str:
        return self.error_message.format(value=self.value)


@dataclass(frozen=True, slots=True, kw_only=True)
class StartsWith:
    prefix: str
    error_message: str = "Value must start with '{prefix}'"

    def get_validator_func(self) -> Callable[[str], bool]:
        def validate(val: str) -> bool:
            return val.startswith(self.prefix)

        return validate

    def get_error_message(self) -> str:
        return self.error_message.format(prefix=self.prefix)


class TestCustomFieldValidator:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            count: Annotated[int, Divisible(value=5)]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"count": 10}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.count == 10

    def test_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            count: Annotated[int, Divisible(value=5)]

        json_file = tmp_path / "config.json"
        content = '{"count": 7}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [count]  Value must be divisible by 5\n   └── FILE '{json_file}', line 1\n       {content}"
        )

    def test_custom_error_message(self, tmp_path: Path):
        @dataclass
        class Config:
            count: Annotated[int, Divisible(value=3, error_message="Must be a multiple of {value}")]

        json_file = tmp_path / "config.json"
        content = '{"count": 7}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [count]  Must be a multiple of 3\n   └── FILE '{json_file}', line 1\n       {content}"
        )


class TestCustomStringValidator:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            url: Annotated[str, StartsWith(prefix="https://")]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"url": "https://example.com"}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.url == "https://example.com"

    def test_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            url: Annotated[str, StartsWith(prefix="https://")]

        json_file = tmp_path / "config.json"
        content = '{"url": "http://example.com"}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [url]  Value must start with 'https://'\n   └── FILE '{json_file}', line 1\n       {content}"
        )


class TestCustomValidatorWithDecorator:
    def test_success(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": 8080}')

        @load(LoadMetadata(file_=json_file))
        @dataclass
        class Config:
            port: Annotated[int, Divisible(value=10)]

        config = Config()
        assert config.port == 8080

    def test_failure(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        content = '{"port": 8081}'
        json_file.write_text(content)

        @load(LoadMetadata(file_=json_file))
        @dataclass
        class Config:
            port: Annotated[int, Divisible(value=10)]

        with pytest.raises(DatureConfigError) as exc_info:
            Config()

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [port]  Value must be divisible by 10\n   └── FILE '{json_file}', line 1\n       {content}"
        )

    def test_direct_instantiation_validates(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        content = '{"port": 8080}'
        json_file.write_text(content)

        @load(LoadMetadata(file_=json_file))
        @dataclass
        class Config:
            port: Annotated[int, Divisible(value=10)]

        with pytest.raises(DatureConfigError) as exc_info:
            Config(port=8081)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [port]  Value must be divisible by 10\n   └── FILE '{json_file}', line 1\n       {content}"
        )


class TestMultipleCustomValidators:
    def test_combined_success(self, tmp_path: Path):
        @dataclass
        class Config:
            count: Annotated[int, Divisible(value=5)]
            url: Annotated[str, StartsWith(prefix="https://")]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"count": 15, "url": "https://example.com"}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.count == 15
        assert result.url == "https://example.com"

    def test_all_fail(self, tmp_path: Path):
        @dataclass
        class Config:
            count: Annotated[int, Divisible(value=5)]
            url: Annotated[str, StartsWith(prefix="https://")]

        json_file = tmp_path / "config.json"
        content = '{"count": 7, "url": "http://example.com"}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 2
        assert str(e) == "Config loading errors (2)"
        assert str(e.exceptions[0]) == (
            f"  [count]  Value must be divisible by 5\n   └── FILE '{json_file}', line 1\n       {content}"
        )
        assert str(e.exceptions[1]) == (
            f"  [url]  Value must start with 'https://'\n   └── FILE '{json_file}', line 1\n       {content}"
        )



from dataclasses import dataclass
from pathlib import Path

import pytest

from dature import LoadMetadata, load
from dature.errors.exceptions import EnvVarExpandError
from dature.field_path import F
from dature.sources_loader.base import BaseLoader
from dature.sources_loader.json_ import JsonLoader
from dature.types import ExpandEnvVarsMode, FileOrStream, JSONValue


class MockLoader(BaseLoader):
    """Mock loader for testing base class functionality."""

    display_name = "mock"

    def __init__(
        self,
        *,
        prefix: str | None = None,
        test_data: JSONValue = None,
        expand_env_vars: ExpandEnvVarsMode = "default",
    ):
        super().__init__(prefix=prefix, expand_env_vars=expand_env_vars)
        self._test_data = test_data or {}

    def _load(self, path: FileOrStream) -> JSONValue:  # noqa: ARG002
        """Return test data."""
        return self._test_data


class TestBaseLoader:
    """Tests for BaseLoader base class."""

    def test_apply_prefix_simple(self):
        """Test applying simple prefix."""
        data = {"app": {"name": "Test", "port": 8080}, "other": "value"}
        loader = MockLoader(prefix="app", test_data=data)

        result = loader._apply_prefix(data)

        assert result == {"name": "Test", "port": 8080}

    def test_apply_prefix_nested(self):
        """Test applying nested prefix with dots."""
        data = {"app": {"database": {"host": "localhost", "port": 5432}}}
        loader = MockLoader(prefix="app.database", test_data=data)

        result = loader._apply_prefix(data)

        assert result == {"host": "localhost", "port": 5432}

    def test_apply_prefix_none(self):
        """Test that None prefix returns original data."""
        data = {"key": "value"}
        loader = MockLoader(test_data=data)

        result = loader._apply_prefix(data)

        assert result == data

    def test_apply_prefix_empty_string(self):
        """Test that empty string prefix returns original data."""
        data = {"key": "value"}
        loader = MockLoader(prefix="", test_data=data)

        result = loader._apply_prefix(data)

        assert result == data

    def test_apply_prefix_nonexistent(self):
        """Test applying nonexistent prefix returns empty dict."""
        data = {"app": {"name": "Test"}}
        loader = MockLoader(prefix="nonexistent", test_data=data)

        result = loader._apply_prefix(data)

        assert result == {}

    def test_apply_prefix_deep_nesting(self):
        """Test applying deeply nested prefix."""
        data = {"a": {"b": {"c": {"d": {"value": "deep"}}}}}
        loader = MockLoader(prefix="a.b.c.d", test_data=data)

        result = loader._apply_prefix(data)

        assert result == {"value": "deep"}

    def test_apply_prefix_invalid_path(self):
        """Test applying prefix with invalid path."""
        data = {"app": "not_a_dict"}
        loader = MockLoader(prefix="app.nested", test_data=data)

        result = loader._apply_prefix(data)

        assert result == {}

    def test_transform_to_dataclass(self):
        """Test transformation of data to dataclass."""

        @dataclass
        class Config:
            name: str
            port: int

        expected_data = Config(name="TestApp", port=8080)
        data = {"name": "TestApp", "port": 8080}
        loader = MockLoader(test_data=data)

        result = loader.transform_to_dataclass(data, Config)

        assert result == expected_data

    def test_transform_to_dataclass_with_nested(self):
        """Test transformation with nested dataclass."""

        @dataclass
        class DatabaseConfig:
            host: str
            port: int

        @dataclass
        class Config:
            database: DatabaseConfig

        expected_data = Config(database=DatabaseConfig(host="localhost", port=5432))
        data = {"database": {"host": "localhost", "port": 5432}}
        loader = MockLoader(test_data=data)

        result = loader.transform_to_dataclass(data, Config)

        assert result == expected_data

    def test_load_raw_and_transform(self):
        """Test load_raw + transform_to_dataclass pipeline."""

        @dataclass
        class Config:
            name: str
            port: int
            debug: bool
            default: str = "value"

        expected_data = Config(name="TestApp", port=8080, debug=True, default="value")
        data = {"app": {"name": "TestApp", "port": 8080, "debug": True}}
        loader = MockLoader(prefix="app", test_data=data)

        load_result = loader.load_raw(Path())
        result = loader.transform_to_dataclass(load_result.data, Config)

        assert result == expected_data

    def test_apply_prefix_with_list(self):
        """Test that apply_prefix returns data as-is when prefix points to non-dict."""
        data = {"items": [1, 2, 3]}
        loader = MockLoader(prefix="items", test_data=data)

        result = loader._apply_prefix(data)

        assert result == [1, 2, 3]


class TestNameStyleMapping:
    def test_lower_snake_to_lower_camel(self, tmp_path: Path):
        @dataclass
        class Config:
            user_name: str
            user_age: int
            is_active: bool

        json_file = tmp_path / "config.json"
        json_file.write_text('{"userName": "John", "userAge": 25, "isActive": true}')

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, name_style="lower_camel"),
            Config,
        )

        assert result.user_name == "John"
        assert result.user_age == 25
        assert result.is_active is True

    def test_lower_camel_to_lower_snake(self, tmp_path: Path):
        @dataclass
        class Config:
            user_name: str
            user_age: int

        json_file = tmp_path / "config.json"
        json_file.write_text('{"user_name": "Alice", "user_age": 30}')

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, name_style="lower_snake"),
            Config,
        )

        assert result.user_name == "Alice"
        assert result.user_age == 30

    def test_upper_camel_pascal_case(self, tmp_path: Path):
        @dataclass
        class Config:
            user_name: str
            total_count: int

        json_file = tmp_path / "config.json"
        json_file.write_text('{"UserName": "Bob", "TotalCount": 100}')

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, name_style="upper_camel"),
            Config,
        )

        assert result.user_name == "Bob"
        assert result.total_count == 100

    def test_lower_kebab_case(self, tmp_path: Path):
        @dataclass
        class Config:
            user_name: str
            api_key: str

        json_file = tmp_path / "config.json"
        json_file.write_text('{"user-name": "Charlie", "api-key": "secret123"}')

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, name_style="lower_kebab"),
            Config,
        )

        assert result.user_name == "Charlie"
        assert result.api_key == "secret123"

    def test_upper_kebab_case(self, tmp_path: Path):
        @dataclass
        class Config:
            user_name: str
            api_key: str

        json_file = tmp_path / "config.json"
        json_file.write_text('{"USER-NAME": "Dave", "API-KEY": "secret456"}')

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, name_style="upper_kebab"),
            Config,
        )

        assert result.user_name == "Dave"
        assert result.api_key == "secret456"

    def test_upper_snake_case(self, tmp_path: Path):
        @dataclass
        class Config:
            user_name: str
            max_retries: int

        json_file = tmp_path / "config.json"
        json_file.write_text('{"USER_NAME": "Eve", "MAX_RETRIES": 3}')

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, name_style="upper_snake"),
            Config,
        )

        assert result.user_name == "Eve"
        assert result.max_retries == 3


class TestFieldMapping:
    def test_simple_field_mapping(self, tmp_path: Path):
        @dataclass
        class Config:
            name: str
            age: int
            active: bool

        json_file = tmp_path / "config.json"
        json_file.write_text('{"fullName": "John Doe", "userAge": 42, "isActive": true}')

        field_mapping = {
            F[Config].name: "fullName",
            F[Config].age: "userAge",
            F[Config].active: "isActive",
        }

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, field_mapping=field_mapping),
            Config,
        )

        assert result.name == "John Doe"
        assert result.age == 42
        assert result.active is True

    def test_partial_field_mapping(self, tmp_path: Path):
        @dataclass
        class Config:
            name: str
            age: int
            city: str

        json_file = tmp_path / "config.json"
        json_file.write_text('{"userName": "Alice", "age": 28, "city": "NYC"}')

        field_mapping = {F[Config].name: "userName"}

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, field_mapping=field_mapping),
            Config,
        )

        assert result.name == "Alice"
        assert result.age == 28
        assert result.city == "NYC"

    def test_combined_name_style_and_field_mapping(self, tmp_path: Path):
        @dataclass
        class Config:
            user_name: str
            user_age: int
            special_field: str

        json_file = tmp_path / "config.json"
        json_file.write_text('{"userName": "Bob", "userAge": 35, "customKey": "special"}')

        field_mapping = {F[Config].special_field: "customKey"}

        result = load(
            LoadMetadata(
                file_=json_file,
                loader=JsonLoader,
                name_style="lower_camel",
                field_mapping=field_mapping,
            ),
            Config,
        )

        assert result.user_name == "Bob"
        assert result.user_age == 35
        assert result.special_field == "special"

    def test_nested_dataclass_with_field_mapping(self, tmp_path: Path):
        @dataclass
        class Address:
            city: str
            street: str

        @dataclass
        class User:
            name: str
            address: Address

        json_file = tmp_path / "config.json"
        json_file.write_text(
            '{"fullName": "Charlie", "location": {"cityName": "LA", "streetName": "Main St"}}',
        )

        field_mapping = {
            F[User].name: "fullName",
            F[User].address: "location",
            F[Address].city: "cityName",
            F[Address].street: "streetName",
        }

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, field_mapping=field_mapping),
            User,
        )

        assert result.name == "Charlie"
        assert result.address.city == "LA"
        assert result.address.street == "Main St"

    def test_tuple_aliases_first_match_wins(self, tmp_path: Path):
        @dataclass
        class Config:
            name: str

        json_file = tmp_path / "config.json"
        json_file.write_text('{"fullName": "Alice"}')

        field_mapping = {F[Config].name: ("fullName", "userName")}

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, field_mapping=field_mapping),
            Config,
        )

        assert result.name == "Alice"

    def test_tuple_aliases_fallback_to_second(self, tmp_path: Path):
        @dataclass
        class Config:
            name: str

        json_file = tmp_path / "config.json"
        json_file.write_text('{"userName": "Bob"}')

        field_mapping = {F[Config].name: ("fullName", "userName")}

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, field_mapping=field_mapping),
            Config,
        )

        assert result.name == "Bob"

    def test_nested_field_path_resolves_owner(self, tmp_path: Path):
        @dataclass
        class Address:
            city: str

        @dataclass
        class User:
            address: Address

        json_file = tmp_path / "config.json"
        json_file.write_text('{"address": {"cityName": "LA"}}')

        field_mapping = {F[User].address.city: "cityName"}

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, field_mapping=field_mapping),
            User,
        )

        assert result.address.city == "LA"

    def test_string_owner_field_mapping(self, tmp_path: Path):
        @dataclass
        class Config:
            name: str

        json_file = tmp_path / "config.json"
        json_file.write_text('{"fullName": "Eve"}')

        field_mapping = {F["Config"].name: "fullName"}

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, field_mapping=field_mapping),
            Config,
        )

        assert result.name == "Eve"

    def test_canonical_name_takes_priority_over_alias(self, tmp_path: Path):
        @dataclass
        class Config:
            name: str

        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "Direct", "fullName": "Alias"}')

        field_mapping = {F[Config].name: "fullName"}

        result = load(
            LoadMetadata(file_=json_file, loader=JsonLoader, field_mapping=field_mapping),
            Config,
        )

        assert result.name == "Direct"

    def test_identity_map_scoped_to_owner_with_name_style(self, tmp_path: Path):
        @dataclass
        class Inner:
            user_name: str
            display_name: str

        @dataclass
        class Config:
            user_name: str
            inner: Inner

        json_file = tmp_path / "config.json"
        json_file.write_text(
            '{"full_name": "Alice", "inner": {"inner_user": "Bob", "displayName": "Bobby"}}',
        )

        field_mapping = {
            F[Config].user_name: "full_name",
            F[Config].inner.user_name: "inner.inner_user",
        }

        result = load(
            LoadMetadata(
                file_=json_file,
                loader=JsonLoader,
                name_style="lower_camel",
                field_mapping=field_mapping,
            ),
            Config,
        )

        assert result.user_name == "Alice"
        assert result.inner.user_name == "Bob"
        assert result.inner.display_name == "Bobby"

    def test_identity_map_flat_json_with_name_style(self, tmp_path: Path):
        @dataclass
        class Inner:
            user_name: str
            display_name: str

        @dataclass
        class Config:
            user_name: str
            inner: Inner

        json_file = tmp_path / "config.json"
        json_file.write_text(
            '{"full_name": "Alice", "inner_user": "Bob", "inner": {"displayName": "Bobby"}}',
        )

        field_mapping = {
            F[Config].user_name: "full_name",
            F[Config].inner.user_name: "inner_user",
        }

        result = load(
            LoadMetadata(
                file_=json_file,
                loader=JsonLoader,
                name_style="lower_camel",
                field_mapping=field_mapping,
            ),
            Config,
        )

        assert result.user_name == "Alice"
        assert result.inner.user_name == "Bob"
        assert result.inner.display_name == "Bobby"


class TestExpandEnvVars:
    def test_default_expands_existing(self, monkeypatch):
        monkeypatch.setenv("DATURE_TEST_HOST", "localhost")
        data = {"host": "$DATURE_TEST_HOST", "port": 8080}
        loader = MockLoader(test_data=data)

        load_result = loader.load_raw(Path())
        result = loader.transform_to_dataclass(load_result.data, dict)

        assert result == {"host": "localhost", "port": 8080}

    def test_default_keeps_missing(self, monkeypatch):
        monkeypatch.delenv("DATURE_MISSING", raising=False)
        data = {"host": "$DATURE_MISSING", "port": 8080}
        loader = MockLoader(test_data=data)

        load_result = loader.load_raw(Path())
        result = loader.transform_to_dataclass(load_result.data, dict)

        assert result == {"host": "$DATURE_MISSING", "port": 8080}

    def test_disabled(self, monkeypatch):
        monkeypatch.setenv("DATURE_TEST_HOST", "localhost")
        data = {"host": "$DATURE_TEST_HOST", "port": 8080}
        loader = MockLoader(test_data=data, expand_env_vars="disabled")

        load_result = loader.load_raw(Path())
        result = loader.transform_to_dataclass(load_result.data, dict)

        assert result == {"host": "$DATURE_TEST_HOST", "port": 8080}

    def test_empty_replaces_missing_with_empty_string(self, monkeypatch):
        monkeypatch.delenv("DATURE_MISSING", raising=False)
        data = {"host": "$DATURE_MISSING", "port": 8080}
        loader = MockLoader(test_data=data, expand_env_vars="empty")

        load_result = loader.load_raw(Path())
        result = loader.transform_to_dataclass(load_result.data, dict)

        assert result == {"host": "", "port": 8080}

    def test_strict_raises_on_missing(self, monkeypatch):
        monkeypatch.delenv("DATURE_MISSING", raising=False)
        data = {"host": "$DATURE_MISSING", "port": 8080}
        loader = MockLoader(test_data=data, expand_env_vars="strict")

        with pytest.raises(EnvVarExpandError):
            loader.load_raw(Path())

    def test_strict_expands_existing(self, monkeypatch):
        monkeypatch.setenv("DATURE_TEST_HOST", "localhost")
        data = {"host": "$DATURE_TEST_HOST", "port": 8080}
        loader = MockLoader(test_data=data, expand_env_vars="strict")

        load_result = loader.load_raw(Path())
        result = loader.transform_to_dataclass(load_result.data, dict)

        assert result == {"host": "localhost", "port": 8080}



"""__post_init__ validation — error example."""

from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, load

SOURCES_DIR = Path(__file__).parent / "sources"


@dataclass
class Config:
    host: str
    port: int
    debug: bool = False

    def __post_init__(self) -> None:
        if self.port < 1 or self.port > 65535:
            msg = f"port must be between 1 and 65535, got {self.port}"
            raise ValueError(msg)

    @property
    def address(self) -> str:
        return f"{self.host}:{self.port}"


try:
    load(LoadMetadata(file_=SOURCES_DIR / "validation_post_init_invalid.yaml"), Config)
except ValueError as exc:
    assert str(exc) == "port must be between 1 and 65535, got 99999"



"""Tests for TypeLoader — custom type loading via LoadMetadata, configure(), and MergeMetadata."""

from collections.abc import Generator
from dataclasses import dataclass
from pathlib import Path

import pytest

from dature import LoadMetadata, MergeMetadata, TypeLoader, configure, load
from dature.config import _ConfigProxy


@dataclass
class Rgb:
    r: int
    g: int
    b: int


def rgb_from_string(value: str) -> Rgb:
    parts = value.split(",")
    return Rgb(r=int(parts[0]), g=int(parts[1]), b=int(parts[2]))


@dataclass
class ConfigWithRgb:
    name: str
    color: Rgb


@pytest.fixture
def _reset_config() -> Generator[None]:
    _ConfigProxy.set_instance(None)
    _ConfigProxy.set_type_loaders(())
    yield
    _ConfigProxy.set_instance(None)
    _ConfigProxy.set_type_loaders(())


@pytest.fixture
def yaml_with_rgb(tmp_path: Path) -> Path:
    p = tmp_path / "rgb.yaml"
    p.write_text("name: test\ncolor: '255,128,0'\n")
    return p


class TestTypeLoadersInLoadMetadata:
    def test_single_source_with_type_loader(self, yaml_with_rgb: Path) -> None:
        result = load(
            LoadMetadata(
                file_=yaml_with_rgb,
                type_loaders=(TypeLoader(type_=Rgb, func=rgb_from_string),),
            ),
            ConfigWithRgb,
        )
        assert result.name == "test"
        assert result.color == Rgb(r=255, g=128, b=0)

    def test_type_loader_overrides_default(self, tmp_path: Path) -> None:
        """TypeLoader for int should override adaptix default."""

        def int_times_two(value: str) -> int:
            return int(value) * 2

        p = tmp_path / "cfg.yaml"
        p.write_text("name: app\ncolor: '10,20,30'\n")

        result = load(
            LoadMetadata(
                file_=p,
                type_loaders=(TypeLoader(type_=Rgb, func=rgb_from_string),),
            ),
            ConfigWithRgb,
        )
        assert result.color == Rgb(r=10, g=20, b=30)


class TestTypeLoadersInConfigure:
    @pytest.mark.usefixtures("_reset_config")
    def test_global_type_loaders_via_configure(self, yaml_with_rgb: Path) -> None:
        configure(
            type_loaders=(TypeLoader(type_=Rgb, func=rgb_from_string),),
        )
        result = load(LoadMetadata(file_=yaml_with_rgb), ConfigWithRgb)
        assert result.color == Rgb(r=255, g=128, b=0)


class TestTypeLoadersInMergeMetadata:
    def test_merge_metadata_type_loaders(self, tmp_path: Path) -> None:
        base = tmp_path / "base.yaml"
        base.write_text("name: base\ncolor: '1,2,3'\n")
        override = tmp_path / "override.yaml"
        override.write_text("name: override\n")

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=base),
                    LoadMetadata(file_=override),
                ),
                type_loaders=(TypeLoader(type_=Rgb, func=rgb_from_string),),
            ),
            ConfigWithRgb,
        )
        assert result.name == "override"
        assert result.color == Rgb(r=1, g=2, b=3)


class TestTypeLoadersMergedFromBoth:
    @pytest.mark.usefixtures("_reset_config")
    def test_per_source_and_global_type_loaders_merge(self, tmp_path: Path) -> None:
        @dataclass
        class TwoCustom:
            color: Rgb
            tag: str

        def tag_upper(value: str) -> str:
            return value.upper()

        configure(
            type_loaders=(TypeLoader(type_=Rgb, func=rgb_from_string),),
        )

        p = tmp_path / "cfg.yaml"
        p.write_text("color: '10,20,30'\ntag: hello\n")

        result = load(
            LoadMetadata(
                file_=p,
                type_loaders=(TypeLoader(type_=str, func=tag_upper),),
            ),
            TwoCustom,
        )
        assert result.color == Rgb(r=10, g=20, b=30)
        assert result.tag == "HELLO"



"""ENV expansion — strict mode on LoadMetadata."""

import os
from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, load

SOURCES_DIR = Path(__file__).parent / "sources"

os.environ["APP_HOST"] = "https://api.example.com"


@dataclass
class Config:
    resolved_url: str
    fallback_url: str


config = load(
    LoadMetadata(file_=SOURCES_DIR / "advanced_env_expansion_strict.yaml", expand_env_vars="strict"),
    Config,
)

assert config.resolved_url == "https://api.example.com/api/v1"
assert config.fallback_url == "postgres://localhost:5432/dev"



from collections.abc import Buffer
from dataclasses import dataclass
from io import BytesIO, RawIOBase, StringIO
from pathlib import Path

import pytest

from dature.field_path import F
from dature.loading.resolver import resolve_loader, resolve_loader_class
from dature.metadata import LoadMetadata
from dature.sources_loader.docker_secrets import DockerSecretsLoader
from dature.sources_loader.env_ import EnvFileLoader, EnvLoader
from dature.sources_loader.ini_ import IniLoader
from dature.sources_loader.json5_ import Json5Loader
from dature.sources_loader.json_ import JsonLoader
from dature.sources_loader.toml_ import Toml11Loader
from dature.sources_loader.yaml_ import Yaml11Loader, Yaml12Loader


class _DummyRawIO(RawIOBase):
    def readinto(self, b: Buffer) -> int:  # noqa: ARG002
        return 0


class TestResolveLoaderClass:
    def test_explicit_loader(self) -> None:
        assert resolve_loader_class(loader=Yaml11Loader, file_="config.json") is Yaml11Loader

    def test_no_file_returns_env(self) -> None:
        assert resolve_loader_class(loader=None, file_=None) is EnvLoader

    @pytest.mark.parametrize(
        ("extension", "expected"),
        [
            (".env", EnvFileLoader),
            (".yaml", Yaml12Loader),
            (".yml", Yaml12Loader),
            (".json", JsonLoader),
            (".json5", Json5Loader),
            (".toml", Toml11Loader),
            (".ini", IniLoader),
            (".cfg", IniLoader),
        ],
    )
    def test_extension_mapping(self, extension: str, expected: type) -> None:
        assert resolve_loader_class(loader=None, file_=f"config{extension}") is expected

    @pytest.mark.parametrize(
        "filename",
        [".env.local", ".env.development", ".env.production"],
    )
    def test_dotenv_patterns(self, filename: str) -> None:
        assert resolve_loader_class(loader=None, file_=filename) is EnvFileLoader

    def test_unknown_extension_raises(self) -> None:
        with pytest.raises(ValueError, match="Cannot determine loader type"):
            resolve_loader_class(loader=None, file_="config.xyz")

    def test_uppercase_extension(self) -> None:
        assert resolve_loader_class(loader=None, file_="config.JSON") is JsonLoader

    def test_env_loader_with_file_raises(self) -> None:
        with pytest.raises(ValueError, match="EnvLoader reads from environment variables") as exc_info:
            resolve_loader_class(loader=EnvLoader, file_="config.json")

        assert str(exc_info.value) == (
            "EnvLoader reads from environment variables and does not use files. "
            "Remove file_ or use a file-based loader instead (e.g. EnvFileLoader)."
        )

    def test_env_file_loader_with_file_allowed(self) -> None:
        assert resolve_loader_class(loader=EnvFileLoader, file_=".env.local") is EnvFileLoader

    def test_directory_returns_docker_secrets(self, tmp_path) -> None:
        assert resolve_loader_class(loader=None, file_=tmp_path) is DockerSecretsLoader


class TestMissingOptionalDependency:
    @pytest.mark.parametrize(
        ("extension", "extra", "blocked_module"),
        [
            (".toml", "toml", "toml_rs"),
            (".yaml", "yaml", "ruamel"),
            (".yml", "yaml", "ruamel"),
            (".json5", "json5", "json5"),
        ],
    )
    def test_missing_extra_raises_helpful_error(
        self,
        extension,
        extra,
        blocked_module,
        block_import,
    ) -> None:
        with block_import(blocked_module):
            with pytest.raises(ImportError) as exc_info:
                resolve_loader_class(loader=None, file_=f"config{extension}")

            assert str(exc_info.value) == (
                f"To use '{extension}' files, install the '{extra}' extra: pip install dature[{extra}]"
            )


class TestResolveLoader:
    def test_returns_correct_loader_type(self) -> None:
        metadata = LoadMetadata(file_="config.json")

        loader = resolve_loader(metadata)

        assert isinstance(loader, JsonLoader)

    def test_passes_prefix(self) -> None:
        metadata = LoadMetadata(prefix="APP_")

        loader = resolve_loader(metadata)

        assert loader._prefix == "APP_"

    def test_passes_name_style(self) -> None:
        metadata = LoadMetadata(file_="config.json", name_style="lower_snake")

        loader = resolve_loader(metadata)

        assert loader._name_style == "lower_snake"

    def test_passes_field_mapping(self) -> None:
        @dataclass
        class Config:
            key: str

        mapping = {F[Config].key: "value"}
        metadata = LoadMetadata(file_="config.json", field_mapping=mapping)

        loader = resolve_loader(metadata)

        assert loader._field_mapping == mapping

    def test_default_metadata_returns_env_loader(self) -> None:
        metadata = LoadMetadata()

        loader = resolve_loader(metadata)

        assert isinstance(loader, EnvLoader)

    def test_env_with_file_path(self, tmp_path: Path) -> None:
        env_file = tmp_path / ".env"
        env_file.write_text("KEY=VALUE")
        metadata = LoadMetadata(file_=env_file)

        loader = resolve_loader(metadata)

        assert isinstance(loader, EnvFileLoader)


class TestFilelikeResolverValidation:
    @pytest.mark.parametrize("stream", [StringIO(), BytesIO(), _DummyRawIO()])
    def test_file_like_without_loader_raises(self, stream) -> None:
        with pytest.raises(TypeError) as exc_info:
            resolve_loader_class(loader=None, file_=stream)

        assert str(exc_info.value) == (
            "Cannot determine loader type for a file-like object. "
            "Please specify loader explicitly (e.g. loader=JsonLoader)."
        )

    @pytest.mark.parametrize("stream", [StringIO(), BytesIO(), _DummyRawIO()])
    def test_file_like_with_env_loader_raises(self, stream) -> None:
        with pytest.raises(ValueError, match="EnvLoader does not support file-like objects") as exc_info:
            resolve_loader_class(loader=EnvLoader, file_=stream)

        assert str(exc_info.value) == (
            "EnvLoader does not support file-like objects. "
            "Use a file-based loader (e.g. JsonLoader, TomlLoader) with file-like objects."
        )

    @pytest.mark.parametrize("stream", [StringIO(), BytesIO(), _DummyRawIO()])
    def test_file_like_with_docker_secrets_loader_raises(self, stream) -> None:
        with pytest.raises(ValueError, match="DockerSecretsLoader does not support file-like objects") as exc_info:
            resolve_loader_class(loader=DockerSecretsLoader, file_=stream)

        assert str(exc_info.value) == (
            "DockerSecretsLoader does not support file-like objects. "
            "Use a file-based loader (e.g. JsonLoader, TomlLoader) with file-like objects."
        )

    @pytest.mark.parametrize("stream", [StringIO(), BytesIO(), _DummyRawIO()])
    def test_file_like_with_explicit_loader_allowed(self, stream) -> None:
        assert resolve_loader_class(loader=JsonLoader, file_=stream) is JsonLoader



"""SecretStr — masks secret values in str/repr."""

from dataclasses import dataclass

from dature.fields.secret_str import SecretStr

secret = SecretStr("my-database-password")


@dataclass
class Config:
    db_password: SecretStr


config = Config(db_password=secret)

assert str(config.db_password) == "**********"
assert repr(config.db_password) == "SecretStr('**********')"
assert config.db_password.get_secret_value() == "my-database-password"
assert len(config.db_password) == 20



from dature.errors.exceptions import LineRange
from dature.path_finders.ini_ import TablePathFinder


class TestIniFindLineRange:
    def test_key_not_confused_by_continuation_line(self):
        content = "[section]\nstr1 = line1\n  x = 1\nx = real\n"
        finder = TablePathFinder(content)

        assert finder.find_line_range(["section", "x"]) == LineRange(start=4, end=4)

    def test_key_with_colon_separator(self):
        content = "[app]\nstr1: line1\n  host: fake\nhost: production\n"
        finder = TablePathFinder(content)

        assert finder.find_line_range(["app", "host"]) == LineRange(start=4, end=4)

    def test_scalar_value(self):
        content = "[section]\nkey = value\n"
        finder = TablePathFinder(content)

        assert finder.find_line_range(["section", "key"]) == LineRange(start=2, end=2)

    def test_continuation_lines(self):
        content = "[section]\nkey = line1\n  line2\n  line3\n"
        finder = TablePathFinder(content)

        assert finder.find_line_range(["section", "key"]) == LineRange(start=2, end=4)

    def test_not_found(self):
        content = "[section]\nkey = value\n"
        finder = TablePathFinder(content)

        assert finder.find_line_range(["section", "missing"]) is None

    def test_no_continuation(self):
        content = "[app]\nhost = localhost\nport = 8080\n"
        finder = TablePathFinder(content)

        assert finder.find_line_range(["app", "host"]) == LineRange(start=2, end=2)



from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import pytest

from dature import LoadMetadata, load
from dature.errors.exceptions import DatureConfigError
from dature.validators.string import MaxLength, MinLength, RegexPattern


class TestMinLength:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            name: Annotated[str, MinLength(value=3)]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "Alice"}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.name == "Alice"

    def test_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            name: Annotated[str, MinLength(value=5)]

        json_file = tmp_path / "config.json"
        content = '{"name": "Bob"}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [name]  Value must have at least 5 characters\n   └── FILE '{json_file}', line 1\n       {content}"
        )


class TestMaxLength:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            name: Annotated[str, MaxLength(value=10)]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "Alice"}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.name == "Alice"

    def test_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            name: Annotated[str, MaxLength(value=5)]

        json_file = tmp_path / "config.json"
        content = '{"name": "Alexander"}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [name]  Value must have at most 5 characters\n   └── FILE '{json_file}', line 1\n       {content}"
        )


class TestRegexPattern:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            email: Annotated[str, RegexPattern(pattern=r"^[\w\.-]+@[\w\.-]+\.\w+$")]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"email": "test@example.com"}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.email == "test@example.com"

    def test_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            email: Annotated[str, RegexPattern(pattern=r"^[\w\.-]+@[\w\.-]+\.\w+$")]

        json_file = tmp_path / "config.json"
        content = '{"email": "invalid-email"}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            "  [email]  Value must match pattern '^[\\w\\.-]+@[\\w\\.-]+\\.\\w+$'\n"
            f"   └── FILE '{json_file}', line 1\n"
            f"       {content}"
        )


class TestCombined:
    def test_combined_string_validators(self, tmp_path: Path):
        @dataclass
        class Config:
            username: Annotated[str, MinLength(value=3), MaxLength(value=20)]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"username": "john_doe"}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.username == "john_doe"

    def test_combined_string_validators_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            username: Annotated[str, MinLength(value=3), MaxLength(value=20)]

        json_file = tmp_path / "config.json"
        content = '{"username": "this_is_a_very_long_username_that_exceeds_limit"}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [username]  Value must have at most 20 characters\n   └── FILE '{json_file}', line 1\n       {content}"
        )



"""RAISE_ON_CONFLICT — raises if the same key has different values."""

from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, MergeMetadata, MergeStrategy, load

SHARED_DIR = Path(__file__).parents[2] / "shared"


@dataclass
class Config:
    host: str
    port: int
    debug: bool


config = load(
    MergeMetadata(
        sources=(
            LoadMetadata(file_=SHARED_DIR / "common_raise_on_conflict_a.yaml"),
            LoadMetadata(file_=SHARED_DIR / "common_raise_on_conflict_b.yaml"),
        ),
        strategy=MergeStrategy.RAISE_ON_CONFLICT,
    ),
    Config,
)

# Disjoint keys — no conflict
assert config.host == "localhost"
assert config.port == 3000
assert config.debug is True



from collections.abc import Callable
from dataclasses import Field
from pathlib import Path
from typing import Any, ClassVar, Protocol, TypeVar

from adaptix import Retort

from dature.errors.exceptions import SourceLocation
from dature.path_finders.base import PathFinder
from dature.types import FileOrStream, JSONValue, LoadRawResult, NestedConflict

_T = TypeVar("_T")


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Field[Any]]]


class ValidatorProtocol(Protocol):
    def get_validator_func(self) -> Callable[..., bool]: ...

    def get_error_message(self) -> str: ...


class LoaderProtocol(Protocol):
    display_name: ClassVar[str]
    display_label: ClassVar[str]
    path_finder_class: type[PathFinder] | None
    retorts: dict[type, Retort]

    def load_raw(self, path: FileOrStream) -> LoadRawResult: ...

    def transform_to_dataclass(self, data: JSONValue, dataclass_: type[_T]) -> _T: ...

    def create_retort(self) -> Retort: ...

    def create_probe_retort(self) -> Retort: ...

    def create_validating_retort(self, dataclass_: type[_T]) -> Retort: ...

    @classmethod
    def resolve_location(
        cls,
        field_path: list[str],
        file_path: Path | None,
        file_content: str | None,
        prefix: str | None,
        split_symbols: str,
        nested_conflict: NestedConflict | None,
    ) -> list[SourceLocation]: ...



"""Debug logging — all loading steps are logged at DEBUG level under the "dature" logger."""

import io
import logging
from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, MergeMetadata, load

log_stream = io.StringIO()
handler = logging.StreamHandler(log_stream)
handler.setLevel(logging.DEBUG)
logging.getLogger("dature").addHandler(handler)
logging.getLogger("dature").setLevel(logging.DEBUG)

SHARED_DIR = Path(__file__).parents[2] / "shared"


@dataclass
class Config:
    host: str
    port: int
    tags: list[str]


config = load(
    MergeMetadata(
        sources=(
            LoadMetadata(file_=SHARED_DIR / "common_defaults.yaml"),
            LoadMetadata(file_=SHARED_DIR / "common_overrides.yaml"),
        ),
    ),
    Config,
)

log_lines = [line for line in log_stream.getvalue().splitlines() if "[Config]" in line]

defaults = str(SHARED_DIR / "common_defaults.yaml")
overrides = str(SHARED_DIR / "common_overrides.yaml")

assert log_lines == [
    f"[Config] Source 0 loaded: loader=yaml1.2, file={defaults}, keys=['host', 'port', 'tags']",
    "[Config] Source 0 raw data: {'host': 'localhost', 'port': 3000, 'tags': ['default']}",
    f"[Config] Source 1 loaded: loader=yaml1.2, file={overrides}, keys=['host', 'port', 'tags']",
    "[Config] Source 1 raw data: {'host': 'production.example.com', 'port': 8080, 'tags': ['web', 'api']}",
    "[Config] Merge step 0 (strategy=last_wins): added=['host', 'port', 'tags'], overwritten=[]",
    "[Config] State after step 0: {'host': 'localhost', 'port': 3000, 'tags': ['default']}",
    "[Config] Merge step 1 (strategy=last_wins): added=[], overwritten=['host', 'port', 'tags']",
    "[Config] State after step 1: {'host': 'production.example.com', 'port': 8080, 'tags': ['web', 'api']}",
    "[Config] Merged result (strategy=last_wins, 2 sources): {'host': 'production.example.com', 'port': 8080, 'tags': ['web', 'api']}",
    f"[Config] Field 'host' = 'production.example.com'  <-- source 1 ({overrides})",
    f"[Config] Field 'port' = 8080  <-- source 1 ({overrides})",
    f"[Config] Field 'tags' = ['web', 'api']  <-- source 1 ({overrides})",
]



from dature.config import configure
from dature.field_path import F
from dature.load_report import get_load_report
from dature.main import load
from dature.metadata import (
    FieldGroup,
    FieldMergeStrategy,
    LoadMetadata,
    MergeMetadata,
    MergeRule,
    MergeStrategy,
    TypeLoader,
)

__all__ = [
    "F",
    "FieldGroup",
    "FieldMergeStrategy",
    "LoadMetadata",
    "MergeMetadata",
    "MergeRule",
    "MergeStrategy",
    "TypeLoader",
    "configure",
    "get_load_report",
    "load",
]



"""prefix — filter ENV keys by prefix."""

import os
from dataclasses import dataclass

from dature import LoadMetadata, load

os.environ["MYAPP_HOST"] = "localhost"
os.environ["MYAPP_PORT"] = "9090"
os.environ["MYAPP_DEBUG"] = "true"


@dataclass
class Config:
    host: str
    port: int
    debug: bool = False


config = load(LoadMetadata(prefix="MYAPP_"), Config)

assert config.host == "localhost"
assert config.port == 9090
assert config.debug is True



import os
import re

from dature.errors.exceptions import EnvVarExpandError, MissingEnvVarError
from dature.types import ExpandEnvVarsMode, JSONValue

# $VAR, ${VAR}, ${VAR:-default}, %VAR%, $$, %%
_VAR_RE = re.compile(
    r"\$\$"  # escaped $$
    r"|%%"  # escaped %%
    r"|\$\{([^}]+)\}"  # ${VAR} or ${VAR:-default}
    r"|\$([A-Za-z_][A-Za-z0-9_]*)"  # $VAR
    r"|%([A-Za-z_][A-Za-z0-9_]*)%",  # %VAR%
)


def _resolve_brace_default(content: str, full: str) -> str:
    separator = ":-"
    idx = content.find(separator)
    if idx != -1:
        var_name = content[:idx]
        fallback = content[idx + len(separator) :]
        value = os.environ.get(var_name)
        if value is not None:
            return value
        return _expand_string_default(fallback)

    value = os.environ.get(content)
    if value is not None:
        return value
    return full


def _resolve_simple_default(var_name: str, full: str) -> str:
    value = os.environ.get(var_name)
    if value is not None:
        return value
    return full


class _EnvExpander:
    def __init__(self, *, mode: ExpandEnvVarsMode, source_text: str) -> None:
        self._mode = mode
        self._source_text = source_text
        self._errors: list[MissingEnvVarError] = []

    @property
    def errors(self) -> list[MissingEnvVarError]:
        return self._errors

    def __call__(self, match: re.Match[str]) -> str:
        full = match.group(0)

        if full == "$$":
            return "$"
        if full == "%%":
            return "%"

        brace_content = match.group(1)
        dollar_name = match.group(2)
        percent_name = match.group(3)

        if brace_content is not None:
            return self._resolve_brace(brace_content, match.start())
        if dollar_name is not None:
            return self._resolve_var(dollar_name, match.start())
        return self._resolve_var(percent_name, match.start())

    def _resolve_brace(self, content: str, position: int) -> str:
        separator = ":-"
        idx = content.find(separator)
        if idx != -1:
            var_name = content[:idx]
            fallback = content[idx + len(separator) :]
            value = os.environ.get(var_name)
            if value is not None:
                return value
            return expand_string(fallback, mode=self._mode)

        return self._resolve_var(content, position)

    def _resolve_var(self, var_name: str, position: int) -> str:
        value = os.environ.get(var_name)
        if value is not None:
            return value

        if self._mode == "strict":
            self._errors.append(
                MissingEnvVarError(
                    var_name=var_name,
                    position=position,
                    source_text=self._source_text,
                ),
            )

        return ""


def expand_string(text: str, *, mode: ExpandEnvVarsMode) -> str:
    if mode == "disabled":
        return text

    if mode == "default":
        return _expand_string_default(text)

    expander = _EnvExpander(mode=mode, source_text=text)
    result = _VAR_RE.sub(expander, text)

    if expander.errors:
        raise EnvVarExpandError(expander.errors)

    return result


def _expand_string_collect(text: str, *, mode: ExpandEnvVarsMode) -> tuple[str, list[MissingEnvVarError]]:
    """Expand string and return (result, errors) without raising."""
    if mode == "disabled":
        return text, []

    if mode == "default":
        return _expand_string_default(text), []

    expander = _EnvExpander(mode=mode, source_text=text)
    result = _VAR_RE.sub(expander, text)
    return result, expander.errors


def _expand_string_default(text: str) -> str:
    def _replace(match: re.Match[str]) -> str:
        full = match.group(0)

        if full == "$$":
            return "$"
        if full == "%%":
            return "%"

        brace_content = match.group(1)
        if brace_content is not None:
            return _resolve_brace_default(brace_content, full)

        var_name = match.group(2) or match.group(3)
        if var_name is not None:
            return _resolve_simple_default(var_name, full)

        return full

    return _VAR_RE.sub(_replace, text)


def expand_env_vars(data: JSONValue, *, mode: ExpandEnvVarsMode) -> JSONValue:
    if mode == "disabled":
        return data

    if mode != "strict":
        return _expand_recursive(data, mode=mode)

    all_errors: list[MissingEnvVarError] = []
    result = _expand_recursive_collect(data, mode=mode, path=[], errors=all_errors)
    if all_errors:
        raise EnvVarExpandError(all_errors)
    return result


def _expand_recursive(data: JSONValue, *, mode: ExpandEnvVarsMode) -> JSONValue:
    if isinstance(data, str):
        return expand_string(data, mode=mode)

    if isinstance(data, dict):
        return {key: _expand_recursive(value, mode=mode) for key, value in data.items()}

    if isinstance(data, list):
        return [_expand_recursive(item, mode=mode) for item in data]

    return data


def _expand_recursive_collect(
    data: JSONValue,
    *,
    mode: ExpandEnvVarsMode,
    path: list[str],
    errors: list[MissingEnvVarError],
) -> JSONValue:
    if isinstance(data, str):
        result, errs = _expand_string_collect(data, mode=mode)
        for err in errs:
            err.field_path = list(path)
        errors.extend(errs)
        return result

    if isinstance(data, dict):
        return {
            key: _expand_recursive_collect(value, mode=mode, path=[*path, key], errors=errors)
            for key, value in data.items()
        }

    if isinstance(data, list):
        return [_expand_recursive_collect(item, mode=mode, path=path, errors=errors) for item in data]

    return data



from collections.abc import Callable
from pathlib import Path
from typing import Any, overload

from dature.config import config
from dature.loading.multi import merge_load_as_function, merge_make_decorator
from dature.loading.resolver import resolve_loader
from dature.loading.single import load_as_function, make_decorator
from dature.metadata import LoadMetadata, MergeMetadata
from dature.protocols import DataclassInstance
from dature.types import FILE_LIKE_TYPES, FileOrStream


@overload
def load[T](
    metadata: LoadMetadata | MergeMetadata | tuple[LoadMetadata, ...] | None,
    /,
    dataclass_: type[T],
    *,
    debug: bool | None = None,
) -> T: ...


@overload
def load(
    metadata: LoadMetadata | MergeMetadata | tuple[LoadMetadata, ...] | None = None,
    /,
    dataclass_: None = None,
    *,
    cache: bool | None = None,
    debug: bool | None = None,
) -> Callable[[type[DataclassInstance]], type[DataclassInstance]]: ...


# --8<-- [start:load]
def load(
    metadata: LoadMetadata | MergeMetadata | tuple[LoadMetadata, ...] | None = None,
    /,
    dataclass_: type[Any] | None = None,
    *,
    cache: bool | None = None,
    debug: bool | None = None,
) -> Any:
    # --8<-- [end:load]
    if cache is None:
        cache = config.loading.cache
    if debug is None:
        debug = config.loading.debug

    if isinstance(metadata, tuple):
        metadata = MergeMetadata(sources=metadata)

    if isinstance(metadata, MergeMetadata):
        merge_type_loaders = (metadata.type_loaders or ()) + config.type_loaders
        if dataclass_ is not None:
            return merge_load_as_function(metadata, dataclass_, debug=debug, type_loaders=merge_type_loaders)
        return merge_make_decorator(metadata, cache=cache, debug=debug, type_loaders=merge_type_loaders)

    if metadata is None:
        metadata = LoadMetadata()

    type_loaders = (metadata.type_loaders or ()) + config.type_loaders
    loader_instance = resolve_loader(
        metadata,
        type_loaders=type_loaders,
        nested_resolve_strategy=config.loading.nested_resolve_strategy,
    )

    file_or_path: FileOrStream
    if isinstance(metadata.file_, FILE_LIKE_TYPES):
        file_or_path = metadata.file_
    elif metadata.file_ is not None:
        file_or_path = Path(metadata.file_)
    else:
        file_or_path = Path()

    if dataclass_ is not None:
        return load_as_function(
            loader_instance=loader_instance,
            file_path=file_or_path,
            dataclass_=dataclass_,
            metadata=metadata,
            debug=debug,
        )

    return make_decorator(
        loader_instance=loader_instance,
        file_path=file_or_path,
        metadata=metadata,
        cache=cache,
        debug=debug,
    )



"""Tests for deep merge algorithms and conflict detection."""

import pytest

from dature.merging.deep_merge import deep_merge, deep_merge_first_wins, deep_merge_last_wins
from dature.metadata import MergeStrategy


class TestDeepMerge:
    @pytest.mark.parametrize(
        ("base", "override", "strategy", "expected"),
        [
            pytest.param(
                {"a": 1, "b": 2},
                {"b": 3, "c": 4},
                MergeStrategy.LAST_WINS,
                {"a": 1, "b": 3, "c": 4},
                id="flat_last_wins",
            ),
            pytest.param(
                {"a": 1, "b": 2},
                {"b": 3, "c": 4},
                MergeStrategy.FIRST_WINS,
                {"a": 1, "b": 2, "c": 4},
                id="flat_first_wins",
            ),
            pytest.param(
                {"db": {"host": "localhost", "port": 5432}},
                {"db": {"host": "prod-host", "name": "mydb"}},
                MergeStrategy.LAST_WINS,
                {"db": {"host": "prod-host", "port": 5432, "name": "mydb"}},
                id="nested_last_wins",
            ),
            pytest.param(
                {"db": {"host": "localhost", "port": 5432}},
                {"db": {"host": "prod-host", "name": "mydb"}},
                MergeStrategy.FIRST_WINS,
                {"db": {"host": "localhost", "port": 5432, "name": "mydb"}},
                id="nested_first_wins",
            ),
            pytest.param(
                {"a": {"b": {"c": 1, "d": 2}}},
                {"a": {"b": {"c": 99, "e": 3}}},
                MergeStrategy.LAST_WINS,
                {"a": {"b": {"c": 99, "d": 2, "e": 3}}},
                id="deeply_nested",
            ),
            pytest.param(
                {"tags": ["a", "b"]},
                {"tags": ["c"]},
                MergeStrategy.LAST_WINS,
                {"tags": ["c"]},
                id="lists_replaced_entirely",
            ),
            pytest.param(
                {},
                {"a": 1},
                MergeStrategy.LAST_WINS,
                {"a": 1},
                id="empty_base",
            ),
            pytest.param(
                {"a": 1},
                {},
                MergeStrategy.LAST_WINS,
                {"a": 1},
                id="empty_override",
            ),
            pytest.param(
                {},
                {},
                MergeStrategy.LAST_WINS,
                {},
                id="both_empty",
            ),
            pytest.param(
                "old",
                "new",
                MergeStrategy.LAST_WINS,
                "new",
                id="scalar_last_wins",
            ),
            pytest.param(
                "old",
                "new",
                MergeStrategy.FIRST_WINS,
                "old",
                id="scalar_first_wins",
            ),
            pytest.param(
                {"a": None},
                {"a": 1},
                MergeStrategy.LAST_WINS,
                {"a": 1},
                id="none_value_last_wins",
            ),
            pytest.param(
                {"a": None},
                {"a": 1},
                MergeStrategy.FIRST_WINS,
                {"a": None},
                id="none_value_first_wins",
            ),
            pytest.param(
                {"a": {"nested": 1}},
                {"a": "scalar"},
                MergeStrategy.LAST_WINS,
                {"a": "scalar"},
                id="dict_vs_scalar_last_wins",
            ),
            pytest.param(
                {"a": "scalar"},
                {"a": {"nested": 1}},
                MergeStrategy.LAST_WINS,
                {"a": {"nested": 1}},
                id="scalar_vs_dict_last_wins",
            ),
            pytest.param(
                {"a": {"nested": 1}},
                {"a": "scalar"},
                MergeStrategy.FIRST_WINS,
                {"a": {"nested": 1}},
                id="dict_vs_scalar_first_wins",
            ),
        ],
    )
    def test_merge(self, base, override, strategy, expected):
        assert deep_merge(base, override, strategy=strategy) == expected

    def test_raise_on_conflict_strategy_raises_value_error(self):
        with pytest.raises(ValueError, match="RAISE_ON_CONFLICT"):
            deep_merge({"a": 1}, {"a": 2}, strategy=MergeStrategy.RAISE_ON_CONFLICT)


class TestDeepMergeLastWins:
    @pytest.mark.parametrize(
        ("base", "override", "expected"),
        [
            pytest.param("old", "new", "new", id="strings"),
            pytest.param(1, 2, 2, id="ints"),
            pytest.param([1], [2], [2], id="lists"),
            pytest.param({"a": 1}, "scalar", "scalar", id="dict_vs_scalar"),
            pytest.param("scalar", {"a": 1}, {"a": 1}, id="scalar_vs_dict"),
        ],
    )
    def test_non_dict_returns_override(self, base, override, expected):
        assert deep_merge_last_wins(base, override) == expected


class TestDeepMergeFirstWins:
    @pytest.mark.parametrize(
        ("base", "override", "expected"),
        [
            pytest.param("old", "new", "old", id="strings"),
            pytest.param(1, 2, 1, id="ints"),
            pytest.param([1], [2], [1], id="lists"),
            pytest.param({"a": 1}, "scalar", {"a": 1}, id="dict_vs_scalar"),
            pytest.param("scalar", {"a": 1}, "scalar", id="scalar_vs_dict"),
        ],
    )
    def test_non_dict_returns_base(self, base, override, expected):
        assert deep_merge_first_wins(base, override) == expected



"""skip_broken_sources — continue loading when a source is missing."""

from dataclasses import dataclass
from pathlib import Path

from dature import LoadMetadata, MergeMetadata, load

SOURCES_DIR = Path(__file__).parent / "sources"
SHARED_DIR = Path(__file__).parents[2] / "shared"


@dataclass
class Config:
    host: str
    port: int
    debug: bool = False


config = load(
    MergeMetadata(
        sources=(
            LoadMetadata(file_=SHARED_DIR / "common_defaults.yaml"),
            LoadMetadata(file_=SOURCES_DIR / "nonexistent.yaml", skip_if_broken=True),
        ),
    ),
    Config,
)

assert config.host == "localhost"
assert config.port == 3000
assert config.debug is False



from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import pytest

from dature import LoadMetadata, load
from dature.errors.exceptions import DatureConfigError
from dature.validators.number import Ge, Gt, Le, Lt


class TestGt:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Gt(value=0)]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"age": 25}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.age == 25

    def test_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Gt(value=18)]

        json_file = tmp_path / "config.json"
        content = '{"age": 18}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [age]  Value must be greater than 18\n   └── FILE '{json_file}', line 1\n       {content}"
        )


class TestGe:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Ge(value=18)]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"age": 18}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.age == 18

    def test_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Ge(value=18)]

        json_file = tmp_path / "config.json"
        content = '{"age": 17}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [age]  Value must be greater than or equal to 18\n   └── FILE '{json_file}', line 1\n       {content}"
        )


class TestLt:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Lt(value=100)]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"age": 99}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.age == 99

    def test_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Lt(value=100)]

        json_file = tmp_path / "config.json"
        content = '{"age": 100}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [age]  Value must be less than 100\n   └── FILE '{json_file}', line 1\n       {content}"
        )


class TestLe:
    def test_success(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Le(value=100)]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"age": 100}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.age == 100

    def test_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Le(value=100)]

        json_file = tmp_path / "config.json"
        content = '{"age": 101}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [age]  Value must be less than or equal to 100\n   └── FILE '{json_file}', line 1\n       {content}"
        )


class TestCombined:
    def test_combined_numeric_validators(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Ge(value=18), Le(value=65)]

        json_file = tmp_path / "config.json"
        json_file.write_text('{"age": 30}')

        metadata = LoadMetadata(file_=json_file)
        result = load(metadata, Config)

        assert result.age == 30

    def test_combined_numeric_validators_failure(self, tmp_path: Path):
        @dataclass
        class Config:
            age: Annotated[int, Ge(value=18), Le(value=65)]

        json_file = tmp_path / "config.json"
        content = '{"age": 70}'
        json_file.write_text(content)

        metadata = LoadMetadata(file_=json_file)

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        e = exc_info.value
        assert len(e.exceptions) == 1
        assert str(e) == "Config loading errors (1)"
        assert str(e.exceptions[0]) == (
            f"  [age]  Value must be less than or equal to 65\n   └── FILE '{json_file}', line 1\n       {content}"
        )



from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Self

from dature.config import config


@dataclass(frozen=True, slots=True)
class LineRange:
    start: int
    end: int

    def __repr__(self) -> str:
        if self.start == self.end:
            return f"line {self.start}"
        return f"line {self.start}-{self.end}"


@dataclass(frozen=True, slots=True)
class SourceLocation:
    display_label: str
    file_path: Path | None
    line_range: LineRange | None
    line_content: list[str] | None
    env_var_name: str | None
    annotation: str | None = None
    env_var_value: str | None = None


def _truncate_line(line: str) -> str:
    max_length = config.error_display.max_line_length
    if len(line) > max_length:
        return line[: max_length - 3] + "..."
    return line


def _format_content_lines(content: list[str]) -> list[str]:
    max_visible = config.error_display.max_visible_lines
    if len(content) > max_visible:
        visible = content[: max_visible - 1]
        lines = [f"       {_truncate_line(line)}" for line in visible]
        lines.append("       ...")
        return lines

    return [f"       {_truncate_line(line)}" for line in content]


def _format_location(loc: SourceLocation, *, connector: str = "└──") -> list[str]:
    lines: list[str] = []
    suffix = f" ({loc.annotation})" if loc.annotation is not None else ""

    if loc.env_var_name is not None and loc.file_path is None:
        main = f"   {connector} {loc.display_label} '{loc.env_var_name}'"
        if loc.env_var_value is not None:
            main += f" = '{loc.env_var_value}'"
        lines.append(main + suffix)
    elif loc.file_path is not None:
        main = f"   {connector} {loc.display_label} '{loc.file_path}'"
        if loc.line_range is not None:
            main += f", {loc.line_range!r}"
        lines.append(main + suffix)
        if loc.line_content is not None:
            lines.extend(_format_content_lines(loc.line_content))

    return lines


class DatureError(Exception):
    """Базовая ошибка dature."""


class FieldLoadError(DatureError):
    def __init__(
        self,
        *,
        field_path: list[str],
        message: str,
        input_value: str | float | bool | None = None,
        locations: list[SourceLocation] | None = None,
    ) -> None:
        self.field_path = field_path
        self.message = message
        self.input_value = input_value
        self.locations = locations or []
        super().__init__(self._format())

    def _format(self) -> str:
        path_str = ".".join(self.field_path)
        if not path_str:
            path_str = "<root>"
        lines = [f"  [{path_str}]  {self.message}"]
        for i, loc in enumerate(self.locations):
            connector = "└──" if i == len(self.locations) - 1 else "├──"
            lines.extend(_format_location(loc, connector=connector))
        return "\n".join(lines)


class MergeConflictFieldError(DatureError):
    def __init__(
        self,
        *,
        field_path: list[str],
        message: str,
        locations: list[SourceLocation],
    ) -> None:
        self.field_path = field_path
        self.message = message
        self.locations = locations
        super().__init__(self._format())

    def _format(self) -> str:
        path_str = ".".join(self.field_path)
        if not path_str:
            path_str = "<root>"
        lines = [f"  [{path_str}]  {self.message}"]
        for loc in self.locations:
            lines.extend(_format_location(loc))
        return "\n".join(lines)


class SourceLoadError(DatureError):
    def __init__(
        self,
        *,
        message: str,
        location: SourceLocation | None = None,
    ) -> None:
        self.message = message
        self.location = location
        super().__init__(message)


class MissingEnvVarError(DatureError):
    def __init__(
        self,
        *,
        var_name: str,
        position: int,
        source_text: str,
        field_path: list[str] | None = None,
        location: SourceLocation | None = None,
    ) -> None:
        self.var_name = var_name
        self.position = position
        self.source_text = source_text
        self.field_path = field_path or []
        self.location = location
        super().__init__(
            f"Environment variable '{var_name}' is not set (position {position} in '{source_text}')",
        )


class DatureConfigError(ExceptionGroup[DatureError]):
    dataclass_name: str

    def __new__(
        cls,
        dataclass_name: str,
        errors: Sequence[DatureError],
    ) -> Self:
        obj = super().__new__(
            cls,
            f"{dataclass_name} loading errors ({len(errors)})",
            errors,
        )
        obj.dataclass_name = dataclass_name
        return obj

    def __init__(
        self,
        dataclass_name: str,
        errors: Sequence[DatureError],
    ) -> None:
        pass

    def derive(self, excs: Sequence[DatureError], /) -> Self:  # type: ignore[override]
        return self.__class__(self.dataclass_name, list(excs))

    def __str__(self) -> str:
        return f"{self.dataclass_name} loading errors ({len(self.exceptions)})"


class EnvVarExpandError(DatureConfigError):
    def __new__(
        cls,
        errors: Sequence[MissingEnvVarError],
        *,
        dataclass_name: str = "",
    ) -> Self:
        obj = super().__new__(cls, dataclass_name, errors)
        obj.dataclass_name = dataclass_name
        return obj

    def __init__(
        self,
        errors: Sequence[MissingEnvVarError],
        *,
        dataclass_name: str = "",
    ) -> None:
        pass

    def derive(self, excs: Sequence[MissingEnvVarError], /) -> Self:  # type: ignore[override]
        return self.__class__(list(excs), dataclass_name=self.dataclass_name)

    def __str__(self) -> str:
        if self.dataclass_name:
            header = f"{self.dataclass_name} env expand errors ({len(self.exceptions)})"
        else:
            header = f"Missing environment variables ({len(self.exceptions)})"
        lines: list[str] = [header, ""]

        for err in self.exceptions:
            if not isinstance(err, MissingEnvVarError):
                continue
            if err.field_path:
                path_str = ".".join(err.field_path)
            else:
                path_str = "<root>"
            lines.append(f"  [{path_str}]  Missing environment variable '{err.var_name}'")
            if err.location is not None:
                lines.extend(_format_location(err.location))
            lines.append("")

        return "\n".join(lines)


class MergeConflictError(DatureConfigError):
    def __new__(
        cls,
        dataclass_name: str,
        errors: Sequence[MergeConflictFieldError],
    ) -> Self:
        return super().__new__(cls, dataclass_name, errors)

    def __str__(self) -> str:
        lines: list[str] = []
        lines.append(f"{self.dataclass_name} merge conflicts ({len(self.exceptions)})")
        lines.append("")

        for exc in self.exceptions:
            if isinstance(exc, MergeConflictFieldError):
                path_str = ".".join(exc.field_path)
                if not path_str:
                    path_str = "<root>"
                lines.append(f"  [{path_str}]  {exc.message}")
                for loc in exc.locations:
                    lines.extend(_format_location(loc))
                lines.append("")
            else:
                lines.append(f"  {exc}")
                lines.append("")

        return "\n".join(lines)


class FieldGroupViolationError(DatureError):
    def __init__(
        self,
        *,
        group_fields: tuple[str, ...],
        changed_fields: tuple[str, ...],
        unchanged_fields: tuple[str, ...],
        changed_sources: tuple[str, ...],
        unchanged_sources: tuple[str, ...],
        source_index: int,
    ) -> None:
        self.group_fields = group_fields
        self.changed_fields = changed_fields
        self.unchanged_fields = unchanged_fields
        self.changed_sources = changed_sources
        self.unchanged_sources = unchanged_sources
        self.source_index = source_index
        super().__init__(self._format())

    def _format(self) -> str:
        group_str = ", ".join(self.group_fields)
        changed_pairs = zip(self.changed_fields, self.changed_sources, strict=True)
        changed_parts = [f"{field} (from source {src})" for field, src in changed_pairs]
        unchanged_pairs = zip(self.unchanged_fields, self.unchanged_sources, strict=True)
        unchanged_parts = [f"{field} (from source {src})" for field, src in unchanged_pairs]
        lines = [
            f"  Field group ({group_str}) partially overridden in source {self.source_index}",
            f"    changed:   {', '.join(changed_parts)}",
            f"    unchanged: {', '.join(unchanged_parts)}",
        ]
        return "\n".join(lines)


class FieldGroupError(DatureConfigError):
    def __new__(
        cls,
        dataclass_name: str,
        errors: Sequence[FieldGroupViolationError],
    ) -> Self:
        return super().__new__(cls, dataclass_name, errors)

    def __str__(self) -> str:
        lines: list[str] = []
        lines.append(f"{self.dataclass_name} field group errors ({len(self.exceptions)})")
        lines.append("")

        for exc in self.exceptions:
            if isinstance(exc, FieldGroupViolationError):
                group_str = ", ".join(exc.group_fields)
                changed_pairs = zip(exc.changed_fields, exc.changed_sources, strict=True)
                changed_parts = [f"{field} (from source {src})" for field, src in changed_pairs]
                unchanged_pairs = zip(exc.unchanged_fields, exc.unchanged_sources, strict=True)
                unchanged_parts = [f"{field} (from source {src})" for field, src in unchanged_pairs]
                lines.append(
                    f"  Field group ({group_str}) partially overridden in source {exc.source_index}",
                )
                lines.append(f"    changed:   {', '.join(changed_parts)}")
                lines.append(f"    unchanged: {', '.join(unchanged_parts)}")
                lines.append("")
            else:
                lines.append(f"  {exc}")
                lines.append("")

        return "\n".join(lines)



from dature.fields.secret_str import SecretStr


class TestSecretStr:
    def test_get_secret_value(self) -> None:
        secret = SecretStr("my_secret")
        assert secret.get_secret_value() == "my_secret"

    def test_str_masks_value(self) -> None:
        secret = SecretStr("my_secret")
        assert str(secret) == "**********"

    def test_repr_masks_value(self) -> None:
        secret = SecretStr("my_secret")
        assert repr(secret) == "SecretStr('**********')"

    def test_eq(self) -> None:
        assert SecretStr("abc") == SecretStr("abc")
        assert SecretStr("abc") != SecretStr("xyz")

    def test_eq_different_type(self) -> None:
        assert SecretStr("abc") != "abc"

    def test_hash(self) -> None:
        assert hash(SecretStr("abc")) == hash(SecretStr("abc"))
        assert hash(SecretStr("abc")) != hash(SecretStr("xyz"))

    def test_len(self) -> None:
        assert len(SecretStr("hello")) == 5
        assert len(SecretStr("")) == 0



"""Field groups — ensure related fields change together."""

from dataclasses import dataclass
from pathlib import Path

from dature import F, FieldGroup, LoadMetadata, MergeMetadata, load

SHARED_DIR = Path(__file__).parents[2] / "shared"


@dataclass
class Config:
    host: str
    port: int
    debug: bool
    user: str
    password: str


config = load(
    MergeMetadata(
        sources=(
            LoadMetadata(file_=SHARED_DIR / "common_field_groups_defaults.yaml"),
            LoadMetadata(file_=SHARED_DIR / "common_field_groups_overrides.yaml"),
        ),
        field_groups=(FieldGroup(F[Config].host, F[Config].port),),
    ),
    Config,
)

assert config.host == "production.example.com"
assert config.port == 8080
assert config.user == "admin"
assert config.password == "secret"



"""Tests for loading/multi.py — multi-source config loading."""

from dataclasses import dataclass
from enum import Flag
from pathlib import Path
from textwrap import dedent
from typing import Annotated

import pytest

from dature import LoadMetadata, MergeMetadata, MergeStrategy, load
from dature.errors.exceptions import DatureConfigError, MergeConflictError
from dature.validators.number import Ge


class TestMergeLoadAsFunction:
    def test_two_json_sources_last_wins(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=defaults),
                    LoadMetadata(file_=overrides),
                ),
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 8080

    def test_two_json_sources_first_wins(self, tmp_path: Path):
        first = tmp_path / "first.json"
        first.write_text('{"host": "first-host", "port": 3000}')

        second = tmp_path / "second.json"
        second.write_text('{"host": "second-host", "port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=first),
                    LoadMetadata(file_=second),
                ),
                strategy=MergeStrategy.FIRST_WINS,
            ),
            Config,
        )

        assert result.host == "first-host"
        assert result.port == 3000

    def test_partial_sources(self, tmp_path: Path):
        file_a = tmp_path / "a.json"
        file_a.write_text('{"host": "myhost"}')

        file_b = tmp_path / "b.json"
        file_b.write_text('{"port": 9090}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=file_a),
                    LoadMetadata(file_=file_b),
                ),
            ),
            Config,
        )

        assert result.host == "myhost"
        assert result.port == 9090

    def test_nested_dataclass(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"database": {"host": "localhost", "port": 5432}}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"database": {"host": "prod-host"}}')

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=defaults),
                    LoadMetadata(file_=overrides),
                ),
            ),
            Config,
        )

        assert result.database.host == "prod-host"
        assert result.database.port == 5432

    def test_three_sources(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{"host": "a-host", "port": 1000, "debug": false}')

        b = tmp_path / "b.json"
        b.write_text('{"port": 2000}')

        c = tmp_path / "c.json"
        c.write_text('{"debug": true}')

        @dataclass
        class Config:
            host: str
            port: int
            debug: bool

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=a),
                    LoadMetadata(file_=b),
                    LoadMetadata(file_=c),
                ),
            ),
            Config,
        )

        assert result.host == "a-host"
        assert result.port == 2000
        assert result.debug is True

    def test_tuple_shorthand(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            (
                LoadMetadata(file_=defaults),
                LoadMetadata(file_=overrides),
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 8080

    def test_json_and_env(self, tmp_path: Path, monkeypatch):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        monkeypatch.setenv("APP_PORT", "9090")
        monkeypatch.setenv("APP_HOST", "env-host")

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=defaults),
                    LoadMetadata(prefix="APP_"),
                ),
            ),
            Config,
        )

        assert result.host == "env-host"
        assert result.port == 9090

    def test_json_and_env_missing_field_error(self, tmp_path: Path, monkeypatch):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost"}')

        monkeypatch.delenv("APP_PORT", raising=False)
        monkeypatch.delenv("APP_HOST", raising=False)

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=defaults),
                        LoadMetadata(prefix="APP_"),
                    ),
                ),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == ("  [port]  Missing required field\n   └── ENV 'APP_PORT'")

    def test_missing_field_in_all_sources(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{"host": "localhost"}')

        b = tmp_path / "b.json"
        b.write_text("{}")

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=a),
                        LoadMetadata(file_=b),
                    ),
                ),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == (f"  [port]  Missing required field\n   └── FILE '{b}'")

    def test_backward_compat_single_load_metadata(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test", "port": 8080}')

        @dataclass
        class Config:
            name: str
            port: int

        result = load(LoadMetadata(file_=json_file), Config)

        assert result.name == "test"
        assert result.port == 8080

    def test_backward_compat_none_metadata(self, monkeypatch):
        monkeypatch.setenv("MY_VAR", "from_env")

        @dataclass
        class Config:
            my_var: str

        result = load(None, Config)

        assert result.my_var == "from_env"


class TestMergeAsDecorator:
    def test_decorator_with_merge(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"port": 9090}')

        meta = MergeMetadata(
            sources=(
                LoadMetadata(file_=defaults),
                LoadMetadata(file_=overrides),
            ),
        )

        @load(meta)
        @dataclass
        class Config:
            host: str
            port: int

        config = Config()
        assert config.host == "localhost"
        assert config.port == 9090

    def test_decorator_cache(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "original", "port": 3000}')

        meta = MergeMetadata(sources=(LoadMetadata(file_=defaults),))

        @load(meta)
        @dataclass
        class Config:
            host: str
            port: int

        first = Config()
        defaults.write_text('{"host": "updated", "port": 9090}')
        second = Config()

        assert first.host == "original"
        assert second.host == "original"

    def test_decorator_no_cache(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "original", "port": 3000}')

        meta = MergeMetadata(sources=(LoadMetadata(file_=defaults),))

        @load(meta, cache=False)
        @dataclass
        class Config:
            host: str
            port: int

        first = Config()
        defaults.write_text('{"host": "updated", "port": 9090}')
        second = Config()

        assert first.host == "original"
        assert second.host == "updated"

    def test_decorator_with_tuple(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"port": 8080}')

        @load(
            (
                LoadMetadata(file_=defaults),
                LoadMetadata(file_=overrides),
            ),
        )
        @dataclass
        class Config:
            host: str
            port: int

        config = Config()
        assert config.host == "localhost"
        assert config.port == 8080

    def test_decorator_init_override(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        meta = MergeMetadata(sources=(LoadMetadata(file_=defaults),))

        @load(meta)
        @dataclass
        class Config:
            host: str
            port: int

        config = Config(host="overridden")
        assert config.host == "overridden"
        assert config.port == 3000

    def test_decorator_not_dataclass(self):
        meta = MergeMetadata(sources=(LoadMetadata(),))

        with pytest.raises(TypeError, match="must be a dataclass"):

            @load(meta)
            class NotDataclass:
                pass

    def test_decorator_first_wins(self, tmp_path: Path):
        first = tmp_path / "first.json"
        first.write_text('{"host": "first-host", "port": 1000}')

        second = tmp_path / "second.json"
        second.write_text('{"host": "second-host", "port": 2000}')

        meta = MergeMetadata(
            sources=(
                LoadMetadata(file_=first),
                LoadMetadata(file_=second),
            ),
            strategy=MergeStrategy.FIRST_WINS,
        )

        @load(meta)
        @dataclass
        class Config:
            host: str
            port: int

        config = Config()
        assert config.host == "first-host"
        assert config.port == 1000


class TestRaiseOnConflict:
    def test_raises_on_scalar_conflict(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{\n  "host": "host-a",\n  "port": 3000\n}')

        b = tmp_path / "b.json"
        b.write_text('{\n  "host": "host-b"\n}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(MergeConflictError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=a),
                        LoadMetadata(file_=b),
                    ),
                    strategy=MergeStrategy.RAISE_ON_CONFLICT,
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config merge conflicts (1)

              [host]  Conflicting values in multiple sources
               └── FILE '{a}', line 2
                   "host": "host-a",
               └── FILE '{b}', line 2
                   "host": "host-b"
            """)

    def test_no_conflict_disjoint_keys(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{"host": "localhost"}')

        b = tmp_path / "b.json"
        b.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=a),
                    LoadMetadata(file_=b),
                ),
                strategy=MergeStrategy.RAISE_ON_CONFLICT,
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 8080

    def test_no_conflict_same_values(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{"host": "same", "port": 3000}')

        b = tmp_path / "b.json"
        b.write_text('{"host": "same", "port": 3000}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=a),
                    LoadMetadata(file_=b),
                ),
                strategy=MergeStrategy.RAISE_ON_CONFLICT,
            ),
            Config,
        )

        assert result.host == "same"
        assert result.port == 3000

    def test_nested_conflict(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{\n  "database": {\n    "host": "a-host",\n    "port": 5432\n  }\n}')

        b = tmp_path / "b.json"
        b.write_text('{\n  "database": {\n    "host": "b-host"\n  }\n}')

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database

        with pytest.raises(MergeConflictError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=a),
                        LoadMetadata(file_=b),
                    ),
                    strategy=MergeStrategy.RAISE_ON_CONFLICT,
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config merge conflicts (1)

              [database.host]  Conflicting values in multiple sources
               └── FILE '{a}', line 3
                   "host": "a-host",
               └── FILE '{b}', line 3
                   "host": "b-host"
            """)

    def test_conflict_error_message_format(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{\n  "host": "a-host"\n}')

        b = tmp_path / "b.json"
        b.write_text('{\n  "host": "b-host"\n}')

        @dataclass
        class Config:
            host: str

        with pytest.raises(MergeConflictError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=a),
                        LoadMetadata(file_=b),
                    ),
                    strategy=MergeStrategy.RAISE_ON_CONFLICT,
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config merge conflicts (1)

              [host]  Conflicting values in multiple sources
               └── FILE '{a}', line 2
                   "host": "a-host"
               └── FILE '{b}', line 2
                   "host": "b-host"
            """)

    def test_conflict_with_env_source(self, tmp_path: Path, monkeypatch):
        a = tmp_path / "a.json"
        a.write_text('{\n  "host": "json-host",\n  "port": 3000\n}')

        monkeypatch.setenv("APP_HOST", "env-host")

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(MergeConflictError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=a),
                        LoadMetadata(prefix="APP_"),
                    ),
                    strategy=MergeStrategy.RAISE_ON_CONFLICT,
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config merge conflicts (1)

              [host]  Conflicting values in multiple sources
               └── FILE '{a}', line 2
                   "host": "json-host",
               └── ENV 'APP_HOST'
            """)

    def test_multiple_conflicts(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{\n  "host": "a-host",\n  "port": 1000\n}')

        b = tmp_path / "b.json"
        b.write_text('{\n  "host": "b-host",\n  "port": 2000\n}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(MergeConflictError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=a),
                        LoadMetadata(file_=b),
                    ),
                    strategy=MergeStrategy.RAISE_ON_CONFLICT,
                ),
                Config,
            )

        assert len(exc_info.value.exceptions) == 2
        assert str(exc_info.value) == dedent(f"""\
            Config merge conflicts (2)

              [host]  Conflicting values in multiple sources
               └── FILE '{a}', line 2
                   "host": "a-host",
               └── FILE '{b}', line 2
                   "host": "b-host",

              [port]  Conflicting values in multiple sources
               └── FILE '{a}', line 3
                   "port": 1000
               └── FILE '{b}', line 3
                   "port": 2000
            """)

    def test_is_subclass_of_dature_config_error(self):
        assert issubclass(MergeConflictError, DatureConfigError)


class TestMergeWithYamlAndEnvFile:
    def test_yaml_and_env_file(self, tmp_path: Path):
        yaml_file = tmp_path / "defaults.yaml"
        yaml_file.write_text("host: localhost\nport: 3000\n")

        env_file = tmp_path / "overrides.env"
        env_file.write_text("PORT=9090\n")

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=yaml_file),
                    LoadMetadata(file_=env_file),
                ),
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 9090


class _Permission(Flag):
    READ = 1
    WRITE = 2
    EXECUTE = 4


class TestCoerceFlagFieldsMergeMode:
    def test_flag_from_env_file_merge(self, tmp_path: Path):
        json_file = tmp_path / "defaults.json"
        json_file.write_text('{"name": "app"}')

        env_file = tmp_path / "overrides.env"
        env_file.write_text("PERMS=3\n")

        @dataclass
        class Config:
            name: str
            perms: _Permission

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=json_file),
                    LoadMetadata(file_=env_file),
                ),
            ),
            Config,
        )

        assert result.perms == _Permission.READ | _Permission.WRITE

    def test_flag_from_env_vars_merge(self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
        json_file = tmp_path / "defaults.json"
        json_file.write_text('{"name": "app"}')

        monkeypatch.setenv("APP_PERMS", "5")

        @dataclass
        class Config:
            name: str
            perms: _Permission

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=json_file),
                    LoadMetadata(prefix="APP_"),
                ),
            ),
            Config,
        )

        assert result.perms == _Permission.READ | _Permission.EXECUTE

    def test_flag_from_json_merge_as_int(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{"name": "app"}')

        b = tmp_path / "b.json"
        b.write_text('{"perms": 7}')

        @dataclass
        class Config:
            name: str
            perms: _Permission

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=a),
                    LoadMetadata(file_=b),
                ),
            ),
            Config,
        )

        assert result.perms == _Permission.READ | _Permission.WRITE | _Permission.EXECUTE

    def test_flag_decorator_merge_from_env_file(self, tmp_path: Path):
        json_file = tmp_path / "defaults.json"
        json_file.write_text('{"name": "app"}')

        env_file = tmp_path / "overrides.env"
        env_file.write_text("PERMS=6\n")

        @dataclass
        class Config:
            name: str
            perms: _Permission

        meta = MergeMetadata(
            sources=(
                LoadMetadata(file_=json_file),
                LoadMetadata(file_=env_file),
            ),
        )

        @load(meta)
        @dataclass
        class MergedConfig:
            name: str
            perms: _Permission

        config = MergedConfig()
        assert config.perms == _Permission.WRITE | _Permission.EXECUTE


class TestFirstFound:
    def test_uses_first_source(self, tmp_path: Path):
        first = tmp_path / "first.yaml"
        first.write_text("host: first-host\nport: 1000\n")

        second = tmp_path / "second.yaml"
        second.write_text("host: second-host\nport: 2000\n")

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=first),
                    LoadMetadata(file_=second),
                ),
                strategy=MergeStrategy.FIRST_FOUND,
            ),
            Config,
        )

        assert result.host == "first-host"
        assert result.port == 1000

    def test_skips_missing_file(self, tmp_path: Path):
        missing = tmp_path / "missing.yaml"
        fallback = tmp_path / "fallback.yaml"
        fallback.write_text("host: fallback-host\nport: 3000\n")

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=missing),
                    LoadMetadata(file_=fallback),
                ),
                strategy=MergeStrategy.FIRST_FOUND,
            ),
            Config,
        )

        assert result.host == "fallback-host"
        assert result.port == 3000

    def test_skips_broken_file(self, tmp_path: Path):
        broken = tmp_path / "broken.yaml"
        broken.write_text(": invalid: yaml: [")

        fallback = tmp_path / "fallback.yaml"
        fallback.write_text("host: fallback-host\nport: 4000\n")

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=broken),
                    LoadMetadata(file_=fallback),
                ),
                strategy=MergeStrategy.FIRST_FOUND,
            ),
            Config,
        )

        assert result.host == "fallback-host"
        assert result.port == 4000

    def test_all_broken_raises(self, tmp_path: Path):
        missing1 = tmp_path / "missing1.yaml"
        missing2 = tmp_path / "missing2.yaml"

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=missing1),
                        LoadMetadata(file_=missing2),
                    ),
                    strategy=MergeStrategy.FIRST_FOUND,
                ),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == "All 2 source(s) failed to load"

    def test_does_not_merge(self, tmp_path: Path):
        partial = tmp_path / "partial.yaml"
        partial.write_text("host: partial-host\n")

        full = tmp_path / "full.yaml"
        full.write_text("host: full-host\nport: 5000\n")

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=partial),
                        LoadMetadata(file_=full),
                    ),
                    strategy=MergeStrategy.FIRST_FOUND,
                ),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == (f"  [port]  Missing required field\n   └── FILE '{partial}'")

    def test_type_error_not_skipped(self, tmp_path: Path):
        bad_type = tmp_path / "bad_type.yaml"
        bad_type.write_text("host: valid-host\nport: not_a_number\n")

        fallback = tmp_path / "fallback.yaml"
        fallback.write_text("host: fallback-host\nport: 6000\n")

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=bad_type),
                        LoadMetadata(file_=fallback),
                    ),
                    strategy=MergeStrategy.FIRST_FOUND,
                ),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == (
            f"  [port]  invalid literal for int() with base 10: 'not_a_number'\n"
            f"   └── FILE '{bad_type}', line 2\n"
            f"       port: not_a_number"
        )

    def test_validation_error_references_correct_source(self, tmp_path: Path):
        first = tmp_path / "first.yaml"
        first.write_text("host: first-host\nport: 0\n")

        second = tmp_path / "second.yaml"
        second.write_text("host: second-host\nport: 5000\n")

        @dataclass
        class Config:
            host: str
            port: Annotated[int, Ge(value=1)]

        with pytest.raises(DatureConfigError) as exc_info:
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=first),
                        LoadMetadata(file_=second),
                    ),
                    strategy=MergeStrategy.FIRST_FOUND,
                ),
                Config,
            )

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == (
            f"  [port]  Value must be greater than or equal to 1\n   └── FILE '{first}', line 2\n       port: 0"
        )

    def test_validation_error_references_correct_source_decorator(self, tmp_path: Path):
        first = tmp_path / "first.yaml"
        first.write_text("host: first-host\nport: 0\n")

        second = tmp_path / "second.yaml"
        second.write_text("host: second-host\nport: 5000\n")

        @load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=first),
                    LoadMetadata(file_=second),
                ),
                strategy=MergeStrategy.FIRST_FOUND,
            ),
            cache=False,
        )
        @dataclass
        class Config:
            host: str
            port: Annotated[int, Ge(value=1)]

        with pytest.raises(DatureConfigError) as exc_info:
            Config()

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == "Config loading errors (1)"
        assert str(err.exceptions[0]) == (
            f"  [port]  Value must be greater than or equal to 1\n   └── FILE '{first}', line 2\n       port: 0"
        )
